# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.asha` -- Asynchronous Successive Halving Algorithm
===================================================================

.. module:: asha
   :platform: Unix
   :synopsis: Asynchronous Successive Halving Algorithm

"""
import copy
import hashlib
import logging

import numpy

from orion.algo.base import BaseAlgorithm
from orion.algo.space import Fidelity


logger = logging.getLogger(__name__)


REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
ASHA can only be used if there is one fidelity dimension.
For more information on the configuration and usage of ASHA, see
https://orion.readthedocs.io/en/develop/user/algorithms.html#asha
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(min_resources, max_resources, reduction_factor, num_rungs):
    """Compute the budgets used for ASHA"""
    budgets = numpy.logspace(
        numpy.log(min_resources) / numpy.log(reduction_factor),
        numpy.log(max_resources) / numpy.log(reduction_factor),
        num_rungs, base=reduction_factor)
    budgets = (budgets + 0.5).astype(int)

    for i in range(num_rungs - 1):
        if budgets[i] >= budgets[i + 1]:
            budgets[i + 1] = budgets[i] + 1

    if budgets[-1] > max_resources:
        raise ValueError(BUDGET_ERROR.format(min_resources, max_resources, num_rungs))

    return list(budgets)


class ASHA(BaseAlgorithm):
    """Asynchronous Successive Halving Algorithm

    `A simple and robust hyperparameter tuning algorithm with solid theoretical underpinnings
    that exploits parallelism and aggressive early-stopping.`

    For more information on the algorithm, see original paper at https://arxiv.org/abs/1810.05934.

    Li, Liam, et al. "Massively parallel hyperparameter tuning."
    arXiv preprint arXiv:1810.05934 (2018)

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    grace_period: int
        Deprecated and will be removed in v0.2.0. To set min resources, you must define it in the
        prior `fidelity(low, high, base)`.
    max_resources: int
        Deprecated and will be removed in v0.2.0. To set max resources, you must define it in the
        prior `fidelity(low, high, base)`.
    reduction_factor: int
        Deprecated and will be removed in v0.2.0. To set the reduction factor, you must define the
        base of `fidelity(low, high, base)`.
    num_rungs: int, optional
        Number of rungs for the largest bracket. If not defined, it will be equal to (base + 1) of
        the fidelity dimension. In the original paper,
        num_rungs == log(fidelity.high/fidelity.low) / log(fidelity.base) + 1.
        Default: log(fidelity.high/fidelity.low) / log(fidelity.base) + 1
    num_brackets: int
        Using a grace period that is too small may bias ASHA too strongly towards
        fast converging trials that do not lead to best results at convergence (stagglers). To
        overcome this, you can increase the number of brackets, which increases the amount of
        resource required for optimisation but decreases the bias towards stragglers.
        Default: 1

    """

    def __init__(self, space, seed=None, grace_period=None, max_resources=None,
                 reduction_factor=None, num_rungs=None, num_brackets=1):
        super(ASHA, self).__init__(
            space, seed=seed, max_resources=max_resources, grace_period=grace_period,
            reduction_factor=reduction_factor, num_rungs=num_rungs, num_brackets=num_brackets)

        self.trial_info = {}  # Stores Trial -> Bracket

        try:
            fidelity_index = self.fidelity_index
        except IndexError:
            raise RuntimeError(SPACE_ERROR)

        fidelity_dim = space.values()[fidelity_index]

        if grace_period is not None:
            logger.warning(
                'The argument `grace_period` is deprecated and will be removed in v0.2.0. To set '
                'min resources, you must define ' 'it in the prior `fidelity(low, high, base)`.')
            min_resources = grace_period
        else:
            min_resources = fidelity_dim.low

        if max_resources is not None:
            logger.warning(
                'The argument `max_resources` is deprecated and will be removed in v0.2.0. To set '
                'max resources, you must define ' 'it in the prior `fidelity(low, high, base)`')
            max_resources = max_resources
        else:
            max_resources = fidelity_dim.high

        if reduction_factor is not None:
            logger.warning(
                'The argument `reduction_factor` is deprecated and will be removed in v0.2.0. To '
                'set the reduction factor, you must define the base of `fidelity(low, high, base)`')
            reduction_factor = reduction_factor
        else:
            reduction_factor = fidelity_dim.base

        if reduction_factor < 2:
            raise AttributeError("Reduction factor for ASHA needs to be at least 2.")

        if num_rungs is None:
            num_rungs = int(numpy.log(max_resources / min_resources) /
                            numpy.log(reduction_factor) + 1)

        self.num_rungs = num_rungs

        budgets = compute_budgets(min_resources, max_resources, reduction_factor, num_rungs)

        # Tracks state for new trial add
        if num_brackets > num_rungs:
            logger.warning("The input num_brackets %i is larger than the number of rungs %i, "
                           "set num_brackets as %i", num_brackets, num_rungs, num_rungs)
            num_brackets = num_rungs

        self.brackets = [
            Bracket(self, reduction_factor, budgets[bracket_index:])
            for bracket_index in range(num_brackets)
        ]

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.rng = numpy.random.RandomState(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return {'rng_state': self.rng.get_state()}

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self.seed_rng(0)
        self.rng.set_state(state_dict['rng_state'])

    def suggest(self, num=1):
        """Suggest a `num` of new sets of parameters.

        Promote a trial if possible, otherwise randomly draw samples from the space and
        randomly assign to a bracket.

        :param num: how many sets to be suggested.

        .. note:: New parameters must be compliant with the problem's domain
           `orion.algo.space.Space`.
        """
        if num > 1:
            raise ValueError("ASHA should suggest only one point.")

        for bracket in self.brackets:
            candidate = bracket.update_rungs()

            if candidate:
                logger.debug('Promoting')
                return [candidate]

        point = self._grow_point_for_bottom_rung()
        if not point:
            return None

        sizes = numpy.array([len(b.rungs) for b in self.brackets])
        probs = numpy.e**(sizes - sizes.max())
        probs = numpy.array([prob * int(not bracket.is_filled)
                             for prob, bracket in zip(probs, self.brackets)])
        normalized = probs / probs.sum()
        idx = self.rng.choice(len(self.brackets), p=normalized)

        point[self.fidelity_index] = self.brackets[idx].rungs[0][0]

        logger.debug('Sampling for bracket %s %s', idx, self.brackets[idx])

        return [tuple(point)]

    def _grow_point_for_bottom_rung(self):
        """Sample point for the bottom rung"""
        if all(bracket.is_filled for bracket in self.brackets):
            logger.warning('All brackets are filled.')
            return None

        for _attempt in range(100):
            point = list(self.space.sample(1, seed=tuple(self.rng.randint(0, 1000000, size=3)))[0])
            if self.get_id(point) not in self.trial_info:
                break

        num_sample_trials = 0
        if self.get_id(point) in self.trial_info:
            for bracket in self.brackets:
                num_sample_trials += len(bracket.rungs[0][1])

            if num_sample_trials >= self.space.cardinality:
                logger.warning('The number of unique trials of bottom rungs exceeds the search '
                               'space cardinality %i, ASHA algorithm exits.',
                               self.space.cardinality)
                return None
            else:
                raise RuntimeError(
                    'ASHA keeps sampling already existing points. This should not happen, '
                    'please report this error to https://github.com/Epistimio/orion/issues')

        return point

    def get_id(self, point):
        """Compute a unique hash for a point based on params, but not fidelity level."""
        _point = list(point)
        non_fidelity_dims = _point[0:self.fidelity_index]
        non_fidelity_dims.extend(_point[self.fidelity_index + 1:])

        return hashlib.md5(str(non_fidelity_dims).encode('utf-8')).hexdigest()

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        A simple random sampler though does not take anything into account.
        """
        for point, result in zip(points, results):

            _id = self.get_id(point)
            bracket = self.trial_info.get(_id)

            if not bracket:
                fidelity = point[self.fidelity_index]
                brackets = [bracket for bracket in self.brackets
                            if bracket.rungs[0][0] == fidelity]
                if not brackets:
                    raise ValueError(
                        "No bracket found for point {0} with fidelity {1}".format(_id, fidelity))
                bracket = brackets[0]

            try:
                bracket.register(point, result['objective'])
            except IndexError:
                logger.warning('Point registered to wrong bracket. This is likely due '
                               'to a corrupted database, where trials of different fidelity '
                               'have a wrong timestamps.')
                continue

            if _id not in self.trial_info:
                self.trial_info[_id] = bracket

    @property
    def is_done(self):
        """Return True, if all brackets reached their maximum resources."""
        return all(bracket.is_done for bracket in self.brackets)

    @property
    def fidelity_index(self):
        """Compute the index of the point when fidelity is."""
        def _is_fidelity(dim):
            return (isinstance(dim, Fidelity) or
                    (hasattr(dim, 'original_dimension') and
                     isinstance(dim.original_dimension, Fidelity)))

        return [i for i, dim in enumerate(self.space.values()) if _is_fidelity(dim)][0]


class Bracket():
    """Bracket of rungs for the algorithm ASHA.

    Parameters
    ----------
    asha: `ASHA` algorithm
        The asha algorithm object which this bracket will be part of.
    reduction_factor: int
        The factor by which ASHA promotes trials. If the reduction factor is 4,
        it means the number of trials from one fidelity level to the next one is roughly
        divided by 4, and each fidelity level has 4 times more resources than the prior one.
    budgets: list of int
        Budgets used for each rung

    """

    def __init__(self, asha, reduction_factor, budgets):
        self.asha = asha
        self.reduction_factor = reduction_factor
        self.rungs = [(budget, dict()) for budget in budgets]

        logger.debug('Bracket budgets: %s', str([rung[0] for rung in self.rungs]))

    def register(self, point, objective):
        """Register a point in the corresponding rung"""
        fidelity = point[self.asha.fidelity_index]
        rungs = [rung for budget, rung in self.rungs if budget == fidelity]
        if not rungs:
            budgets = [budget for budget, rung in self.rungs]
            raise IndexError(REGISTRATION_ERROR.format(fidelity=fidelity, budgets=budgets,
                                                       params=point))

        rungs[0][self.asha.get_id(point)] = (objective, point)

    def get_candidate(self, rung_id):
        """Get a candidate for promotion"""
        _, rung = self.rungs[rung_id]
        next_rung = self.rungs[rung_id + 1][1]

        rung = list(sorted((objective, point) for objective, point in rung.values()
                           if objective is not None))
        k = len(rung) // self.reduction_factor
        k = min(k, len(rung))

        for i in range(k):
            point = rung[i][1]
            _id = self.asha.get_id(point)
            if _id not in next_rung:
                return point

        return None

    @property
    def is_done(self):
        """Return True, if the last rung is filled."""
        return len(self.rungs[-1][1])

    @property
    def is_filled(self):
        """Return True, if the penultimate rung is filled."""
        return self.has_rung_filled(len(self.rungs) - 2)

    def has_rung_filled(self, rung_id):
        """Return True, if the rung[rung_id] is filled."""
        n_rungs = len(self.rungs)
        n_trials = len(self.rungs[rung_id][1])
        return n_trials >= self.reduction_factor ** (n_rungs - rung_id - 1)

    def update_rungs(self):
        """Promote the first candidate that is found and return it

        The rungs are iterated over in reversed order, so that high rungs
        are prioritised for promotions. When a candidate is promoted, the loop is broken and
        the method returns the promoted point.

        .. note ::

            All trials are part of the rungs, for any state. Only completed trials
            are eligible for promotion, i.e., only completed trials can be part of top-k.
            Lookup for promotion in rung l + 1 contains trials of any status.

        """
        if self.is_done:
            return None

        for rung_id in range(len(self.rungs) - 2, -1, -1):
            candidate = self.get_candidate(rung_id)
            if candidate:

                # pylint: disable=logging-format-interpolation
                logger.debug(
                    'Promoting {point} from rung {past_rung} with fidelity {past_fidelity} to '
                    'rung {new_rung} with fidelity {new_fidelity}'.format(
                        point=candidate, past_rung=rung_id,
                        past_fidelity=candidate[self.asha.fidelity_index],
                        new_rung=rung_id + 1, new_fidelity=self.rungs[rung_id + 1][0]))

                candidate = list(copy.deepcopy(candidate))
                candidate[self.asha.fidelity_index] = self.rungs[rung_id + 1][0]

                return tuple(candidate)

        return None

    def __repr__(self):
        """Return representation of bracket with fidelity levels"""
        return 'Bracket({})'.format([rung[0] for rung in self.rungs])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.checks.creation` -- Creation stage for database checks
===========================================================================

.. module:: creation
    :platform: Unix
    :synopsis: Checks for the creation of a `Database` object.

"""

from orion.core.io.database import Database
from orion.core.utils.exceptions import CheckError


class CreationStage:
    """The creation stage of the checks."""

    def __init__(self, presence_stage):
        """Create an intance of the stage.

        Parameters
        ----------
        presence_stage: `PresenceStage`
            An instance of the previous stage.

        """
        self.p_stage = presence_stage
        self.instance = None

    def checks(self):
        """Return checks."""
        yield self.check_database_creation

    def check_database_creation(self):
        """Check if database of specified type can be created."""
        database = self.p_stage.db_config
        db_type = database.pop('type')

        try:
            db = Database(of_type=db_type, **database)
        except ValueError as ex:
            raise CheckError(str(ex))

        if not db.is_connected:
            raise CheckError("Database failed to connect after creation.")

        self.instance = db

        return "Success", ""

    def post_stage(self):
        """Print the created database."""
        print('DB instance', self.instance)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db.rm` -- Module running the rm command
============================================================

.. module:: setup
   :platform: Unix
   :synopsis: Delete experiments and trials from the database
"""
import argparse
import logging
import sys

import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.pptree import print_tree
from orion.core.utils.terminal import confirm_name
from orion.storage.base import get_storage


logger = logging.getLogger(__name__)


EXP_RM_MESSAGE = """
All experiments above and their corresponding trials will be deleted.
To select a specific version use --version <VERSION>. Note that all
children of a given version will be deleted. Oríon cannot delete a
parent experiment without deleting the children.
To delete trials only, use --status <STATUS>.

Make sure to stop any worker currently executing one of these experiment.

To proceed, type again the name of the experiment: """


TRIALS_RM_MESSAGE = """
Matching trials of all experiments above will be deleted.
To select a specific version use --version <VERSION>.
Note that trials of all children of a given version will be deleted.

Make sure to stop any worker currently executing one of these experiment.

To proceed, type again the name of the experiment: """


DESCRIPTION = """
Command to delete experiments and trials.

To delete an experiment and its trials, simply give the experiment's name.
$ orion db rm my-exp-name

To delete only trials that are broken, simply add --status broken.
Note that the experiment will not be deleted, only the trials.
$ orion db rm my-exp-name --status broken

Or --status * to delete all trials of the experiment.
$ orion db rm my-exp-name --status *

By default, the last version of the experiment is deleted. Add --version
to select a prior version. Note that all child of the selected version
will be deleted as well. You cannot delete a parent experiment without
deleting the child experiments.
$ orion db rm my-exp-name --version 1
"""


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    rm_parser = parser.add_parser(
        'rm',
        description=DESCRIPTION,
        help='Deletes experiments and trials',
        formatter_class=argparse.RawTextHelpFormatter)

    rm_parser.set_defaults(func=main)

    rm_parser.add_argument(
        'name',
        help='Name of the experiment to delete.')

    rm_parser.add_argument('-c', '--config', type=argparse.FileType('r'),
                           metavar='path-to-config', help="user provided "
                           "orion configuration file")

    rm_parser.add_argument(
        '-v', '--version', type=int, default=None,
        help="specific version of experiment to fetch; "
             "(default: last version matching.)")

    rm_parser.add_argument(
        '-s', '--status',
        help='Remove all trials of the experiment with the given status '
             '(Will not delete the experiment). '
             'Also supports --status=* to delete all trials of a given experiment.')

    rm_parser.add_argument(
        '-f', '--force', action='store_true',
        help='Force delete without asking to enter experiment name twice.')

    return rm_parser


def process_trial_rm(storage, root, status):
    """Delete the matching trials of the given experiment."""
    trials_total = 0
    for node in root:
        if status == '*':
            query = {}
        else:
            query = {'status': status}

        count = storage.delete_trials(uid=node.item.id, where=query)
        logger.debug('%d trials deleted in experiment %s-v%d',
                     count, node.item.name, node.item.version)
        trials_total += count

    print(f'{trials_total} trials deleted')


def process_exp_rm(storage, root):
    """Delete the given experiment node and all its children."""
    trials_total = 0
    exp_total = 0
    for node in root:
        count = storage.delete_trials(uid=node.item.id)
        trials_total += count
        logger.debug('%d trials deleted in experiment %s-v%d',
                     count, node.item.name, node.item.version)
        count = storage.delete_experiment(uid=node.item.id)
        logger.debug('%s experiment %s-v%d deleted',
                     count, node.item.name, node.item.version)
        exp_total += count

    print(f'{trials_total} trials deleted')
    print(f'{exp_total} experiments deleted')


def delete_experiments(storage, root, name, force):
    """Delete matching experiments after user confirmation."""
    confirmed = confirm_name(EXP_RM_MESSAGE, name, force)

    if not confirmed:
        print('Confirmation failed, aborting operation.')
        sys.exit(1)

    process_exp_rm(storage, root)


def delete_trials(storage, root, name, status, force):
    """Delete all matching trials after user confirmation."""
    confirmed = confirm_name(TRIALS_RM_MESSAGE, name, force)

    if not confirmed:
        print('Confirmation failed, aborting operation.')
        sys.exit(1)

    process_trial_rm(storage, root, status)


def main(args):
    """Remove the experiment(s) or trial(s)."""
    config = experiment_builder.get_cmd_config(args)
    experiment_builder.setup_storage(config.get('storage'))

    # Find root experiment
    root = experiment_builder.build_view(name=args['name'],
                                         version=args.get('version', None)).node

    # List all experiments with children
    print_tree(root, nameattr='tree_name')

    storage = get_storage()

    if args['status']:
        delete_trials(storage, root, args['name'], args['status'], args['force'])
    else:
        delete_experiments(storage, root, args['name'], args['force'])



# -*- coding: utf-8 -*-
"""
:mod:`orion.testing.state` -- Mocks Oríon's runtime.
====================================================
.. module:: testing
   :platform: Unix
   :synopsis: Boilerplate to simulate Oríon's runtime and data sources
"""
# pylint: disable=protected-access

import os
import tempfile

import yaml

from orion.core.io import experiment_builder as experiment_builder
from orion.core.utils.singleton import SingletonAlreadyInstantiatedError, update_singletons
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage, Storage


# pylint: disable=no-self-use,protected-access
class BaseOrionState:
    """Setup global variables and singleton for tests.

    It swaps the singleton with `None` at startup and restores them after the tests.
    It also initializes PickleDB as the storage for testing.
    We use PickledDB as our storage mock

    Parameters
    ----------
    experiments: list, optional
        List of experiments to insert into the database
    trials: list, optional
        List of trials to insert into the database
    workers: list, optional
        List of workers to insert into the database
    lies: list, optional
        List of lies to insert into the database
    resources: list, optional
        List of resources to insert into the database
    from_yaml: YAML, optional
        YAML config to apply for this test
    storage: dict, optional
        Configuration of the underlying storage backend

    Examples
    --------
    >>> my_config = {...}
    >>> with OrionState(my_config):
        ...

    """

    # TODO: Fix these singletons to remove Legacy, MongoDB, PickledDB and EphemeralDB.
    singletons = {}
    experiments = []
    trials = []
    resources = []
    workers = []

    def __init__(self, experiments=None, trials=None, workers=None, lies=None, resources=None,
                 from_yaml=None, storage=None):
        if from_yaml is not None:
            with open(from_yaml) as f:
                exp_config = list(yaml.safe_load_all(f))
                experiments = exp_config[0]
                trials = exp_config[1]

        self.tempfile = None
        self.tempfile_path = None
        self.storage_config = _select(storage, _get_default_test_storage())

        self._experiments = _select(experiments, [])
        self._trials = _select(trials, [])
        self._workers = _select(workers, [])
        self._resources = _select(resources, [])
        self._lies = _select(lies, [])

        # In case of track we also store the inserted object
        # so the user can compare in tests the different values
        self.trials = []
        self.experiments = self._experiments
        self.lies = []

    def init(self, config):
        """Initialize environment before testing"""
        self.storage(config)
        self.load_experience_configuration()
        return self

    def get_experiment(self, name, version=None):
        """Make experiment id deterministic"""
        exp = experiment_builder.build(name=name, version=version)
        return exp

    def get_trial(self, index):
        """Return a Trial"""
        return Trial(**self.trials[index])

    def cleanup(self):
        """Cleanup after testing"""
        if self.tempfile is not None:
            os.close(self.tempfile)
        _remove(self.tempfile_path)

    def _set_tables(self):
        self.trials = []
        self.lies = []

        for exp in self._experiments:
            get_storage().create_experiment(exp)

        for trial in self._trials:
            nt = get_storage().register_trial(Trial(**trial))
            self.trials.append(nt.to_dict())

        for lie in self._lies:
            nt = get_storage().register_lie(Trial(**lie))
            self.lies.append(nt.to_dict())

    def load_experience_configuration(self):
        """Load an example database."""
        for i, t_dict in enumerate(self._trials):
            self._trials[i] = Trial(**t_dict).to_dict()

        for i, t_dict in enumerate(self._lies):
            self._lies[i] = Trial(**t_dict).to_dict()

        self._trials.sort(key=lambda obj: int(obj['_id'], 16), reverse=True)

        for i, experiment in enumerate(self._experiments):
            if 'user_script' in experiment['metadata']:
                path = os.path.join(
                    os.path.dirname(__file__),
                    experiment["metadata"]["user_script"])
                experiment["metadata"]["user_script"] = path

            experiment['_id'] = i

        self._set_tables()

    def make_config(self):
        """Iterate over the database configuration and replace ${file}
        by the name of a temporary file
        """
        self.tempfile, self.tempfile_path = tempfile.mkstemp('_orion_test')
        _remove(self.tempfile_path)

        def map_dict(fun, dictionary):
            """Return a dictionary with fun applied to each values"""
            return {k: fun(v) for k, v in dictionary.items()}

        def replace_file(v):
            """Replace `${file}` by a generated temporary file"""
            if isinstance(v, str):
                v = v.replace('${file}', self.tempfile_path)

            if isinstance(v, dict):
                v = map_dict(replace_file, v)

            return v

        return map_dict(replace_file, self.storage_config)

    def __enter__(self):
        """Load a new database state"""
        self.singletons = update_singletons()
        self.cleanup()
        return self.init(self.make_config())

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup database state"""
        self.cleanup()

        update_singletons(self.singletons)

    def storage(self, config=None):
        """Return test storage"""
        if config is None:
            return get_storage()

        try:
            config['of_type'] = config.pop('type')
            db = Storage(**config)
            self.storage_config = config
        except SingletonAlreadyInstantiatedError:
            db = get_storage()

        except KeyError:
            print(self.storage_config)
            raise

        return db


class LegacyOrionState(BaseOrionState):
    """See :func:`~orion.utils.tests.BaseOrionState`"""

    def __init__(self, *args, **kwargs):
        super(LegacyOrionState, self).__init__(*args, **kwargs)
        self.initialized = False

    @property
    def database(self):
        """Retrieve legacy database handle"""
        return get_storage()._db

    def init(self, config):
        """Initialize environment before testing"""
        self.storage(config)
        self.initialized = True

        if hasattr(get_storage(), '_db'):
            self.database.remove('experiments', {})
            self.database.remove('trials', {})

        self.load_experience_configuration()
        return self

    def get_experiment(self, name, version=None):
        """Make experiment id deterministic"""
        exp = experiment_builder.build(name, version=version)
        exp._id = exp.name
        return exp

    def _set_tables(self):
        if self._experiments:
            self.database.write('experiments', self._experiments)
        if self._trials:
            self.database.write('trials', self._trials)
        if self._workers:
            self.database.write('workers', self._workers)
        if self._resources:
            self.database.write('resources', self._resources)
        if self._lies:
            self.database.write('lying_trials', self._lies)

        self.lies = self._lies
        self.trials = self._trials

    def cleanup(self):
        """Cleanup after testing"""
        if self.initialized:
            self.database.remove('experiments', {})
            self.database.remove('trials', {})
            if self.tempfile is not None:
                os.close(self.tempfile)
            _remove(self.tempfile_path)
        self.initialized = False


# We are faking a class constructor here
# pylint: disable=C0103
def OrionState(*args, **kwargs):
    """Build an orion state in function of the storage type"""
    storage = kwargs.get('storage')

    if not storage or storage['type'] == 'legacy':
        return LegacyOrionState(*args, **kwargs)

    return BaseOrionState(*args, **kwargs)


def customized_mutate_example(search_space, old_value, **kwargs):
    """Define a customized mutate function example"""
    multiply_factor = kwargs.pop('multiply_factor', 3.0)
    add_factor = kwargs.pop('add_factor', 1)
    if search_space.type == "real":
        new_value = old_value / multiply_factor
    elif search_space.type == "integer":
        new_value = int(old_value + add_factor)
    else:
        new_value = old_value
    return new_value


def _get_default_test_storage():
    """Return default configuration for the test storage"""
    return {
        'type': 'legacy',
        'database': {
            'type': 'PickledDB',
            'host': '${file}'
        }
    }


def _remove(file_name):
    if file_name is None:
        return

    try:
        os.remove(file_name)
    except FileNotFoundError:
        pass


def _select(lhs, rhs):
    if lhs:
        return lhs
    return rhs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.algo.random`."""
import numpy
import pytest

from orion.algo.random import Random
from orion.algo.space import Integer, Real, Space


@pytest.fixture()
def space():
    """Return an optimization space"""
    space = Space()
    dim1 = Integer('yolo1', 'uniform', -3, 6)
    space.register(dim1)
    dim2 = Real('yolo2', 'norm', 0.9)
    space.register(dim2)

    return space


def test_seeding(space):
    """Verify that seeding makes sampling deterministic"""
    random_search = Random(space)

    random_search.seed_rng(1)
    a = random_search.suggest(1)[0]
    assert not numpy.allclose(a, random_search.suggest(1)[0])

    random_search.seed_rng(1)
    assert numpy.allclose(a, random_search.suggest(1)[0])


def test_set_state(space):
    """Verify that resetting state makes sampling deterministic"""
    random_search = Random(space)

    random_search.seed_rng(1)
    state = random_search.state_dict
    a = random_search.suggest(1)[0]
    assert not numpy.allclose(a, random_search.suggest(1)[0])

    random_search.set_state(state)
    assert numpy.allclose(a, random_search.suggest(1)[0])


def test_suggest_unique():
    """Verify that RandomSearch do not sample duplicates"""
    space = Space()
    space.register(Integer('yolo1', 'uniform', -3, 6))

    random_search = Random(space)

    n_samples = 6
    values = sum(random_search.suggest(n_samples), tuple())
    assert len(values) == n_samples
    assert len(set(values)) == n_samples


def test_suggest_unique_history():
    """Verify that RandomSearch do not sample duplicates based observed points"""
    space = Space()
    space.register(Integer('yolo1', 'uniform', -3, 6))

    random_search = Random(space)

    n_samples = 3
    values = sum(random_search.suggest(n_samples), tuple())
    assert len(values) == n_samples
    assert len(set(values)) == n_samples

    random_search.observe([[value] for value in values], [1] * n_samples)

    n_samples = 3
    new_values = sum(random_search.suggest(n_samples), tuple())
    assert len(new_values) == n_samples
    assert len(set(new_values)) == n_samples
    # No duplicates
    assert (set(new_values) & set(values)) == set()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for demo purposes."""
from collections import defaultdict
import os
import shutil
import subprocess
import tempfile

import numpy
import pytest
import yaml

import orion.core.cli
from orion.core.io.database.ephemeraldb import EphemeralDB
import orion.core.io.experiment_builder as experiment_builder
from orion.core.worker import workon
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_with_default_algo_cli_config_only(database, monkeypatch):
    """Check that random algorithm is used, when no algo is chosen explicitly."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    monkeypatch.setenv('ORION_DB_NAME', 'orion_test')
    monkeypatch.setenv('ORION_DB_ADDRESS', 'mongodb://user:pass@localhost')

    orion.core.cli.main(["hunt", "-n", "default_algo",
                         "--max-trials", "30",
                         "./black_box.py", "-x~uniform(-50, 50)"])

    exp = list(database.experiments.find({'name': 'default_algo'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    assert exp['name'] == 'default_algo'
    assert exp['pool_size'] == 1
    assert exp['max_trials'] == 30
    assert exp['algorithms'] == {'random': {'seed': None}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == ["./black_box.py", '-x~uniform(-50, 50)']


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo(database, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "./black_box.py",
        "-x~uniform(-50, 50, precision=None)",
        "--test-env",
        "--experiment-id", '{exp.id}',
        "--experiment-name", '{exp.name}',
        "--experiment-version", '{exp.version}',
        "--trial-id", '{trial.id}',
        "--working-dir", '{trial.working_dir}']

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    assert exp['name'] == 'voila_voici'
    assert exp['pool_size'] == 1
    assert exp['max_trials'] == 100
    assert exp['algorithms'] == {'gradient_descent': {'learning_rate': 0.1,
                                                      'dx_tolerance': 1e-7}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == user_args
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) <= 15
    assert trials[-1]['status'] == 'completed'
    trials = list(sorted(trials, key=lambda trial: trial['submit_time']))
    for result in trials[-1]['results']:
        assert result['type'] != 'constraint'
        if result['type'] == 'objective':
            assert abs(result['value'] - 23.4) < 1e-6
            assert result['name'] == 'example_objective'
        elif result['type'] == 'gradient':
            res = numpy.asarray(result['value'])
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-7
            assert result['name'] == 'example_gradient'
    params = trials[-1]['params']
    assert len(params) == 1
    assert params[0]['name'] == '/x'
    assert params[0]['type'] == 'real'
    assert (params[0]['value'] - 34.56789) < 1e-5


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_with_script_config(database, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--config", "./orion_config.yaml",
                         "./black_box_w_config.py", "--config", "script_config.yaml"])

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    assert exp['name'] == 'voila_voici'
    assert exp['pool_size'] == 1
    assert exp['max_trials'] == 100
    assert exp['algorithms'] == {'gradient_descent': {'learning_rate': 0.1,
                                                      'dx_tolerance': 1e-7}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == ['./black_box_w_config.py', '--config',
                                            'script_config.yaml']

    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) <= 15
    assert trials[-1]['status'] == 'completed'
    trials = list(sorted(trials, key=lambda trial: trial['submit_time']))
    for result in trials[-1]['results']:
        assert result['type'] != 'constraint'
        if result['type'] == 'objective':
            assert abs(result['value'] - 23.4) < 1e-6
            assert result['name'] == 'example_objective'
        elif result['type'] == 'gradient':
            res = numpy.asarray(result['value'])
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-7
            assert result['name'] == 'example_gradient'
    params = trials[-1]['params']
    assert len(params) == 1
    assert params[0]['name'] == '/x'
    assert params[0]['type'] == 'real'
    assert (params[0]['value'] - 34.56789) < 1e-5


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_with_python_and_script(database, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--config", "./orion_config.yaml",
                         "python", "black_box_w_config.py", "--config", "script_config.yaml"])

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    assert exp['name'] == 'voila_voici'
    assert exp['pool_size'] == 1
    assert exp['max_trials'] == 100
    assert exp['algorithms'] == {'gradient_descent': {'learning_rate': 0.1,
                                                      'dx_tolerance': 1e-7}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == ['python', 'black_box_w_config.py',
                                            '--config', 'script_config.yaml']

    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) <= 15
    assert trials[-1]['status'] == 'completed'
    trials = list(sorted(trials, key=lambda trial: trial['submit_time']))
    for result in trials[-1]['results']:
        assert result['type'] != 'constraint'
        if result['type'] == 'objective':
            assert abs(result['value'] - 23.4) < 1e-6
            assert result['name'] == 'example_objective'
        elif result['type'] == 'gradient':
            res = numpy.asarray(result['value'])
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-7
            assert result['name'] == 'example_gradient'
    params = trials[-1]['params']
    assert len(params) == 1
    assert params[0]['name'] == '/x'
    assert params[0]['type'] == 'real'
    assert (params[0]['value'] - 34.56789) < 1e-5


@pytest.mark.usefixtures("clean_db")
def test_demo_inexecutable_script(database, monkeypatch, capsys):
    """Test error message when user script is not executable."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    script = tempfile.NamedTemporaryFile()
    orion.core.cli.main(["hunt", "--config", "./orion_config.yaml",
                         script.name, "--config", "script_config.yaml"])

    captured = capsys.readouterr().err
    assert 'User script is not executable' in captured


@pytest.mark.usefixtures("clean_db")
def test_demo_two_workers(database, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    processes = []
    for _ in range(2):
        process = subprocess.Popen(["orion", "hunt", "-n", "two_workers_demo",
                                    "--config", "./orion_config_random.yaml",
                                    "--max-trials", "100",
                                    "./black_box.py", "-x~norm(34, 3)"])
        processes.append(process)

    for process in processes:
        rcode = process.wait()
        assert rcode == 0

    exp = list(database.experiments.find({'name': 'two_workers_demo'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    assert exp['name'] == 'two_workers_demo'
    assert exp['pool_size'] == 2
    assert exp['max_trials'] == 100
    assert exp['algorithms'] == {'random': {'seed': None}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == ['./black_box.py', '-x~norm(34, 3)']

    trials = list(database.trials.find({'experiment': exp_id}))
    status = defaultdict(int)
    for trial in trials:
        status[trial['status']] += 1
    assert 100 <= status['completed'] <= 101
    assert status['new'] < 5
    params = trials[-1]['params']
    assert len(params) == 1
    assert params[0]['name'] == '/x'
    assert params[0]['type'] == 'real'


def test_workon():
    """Test scenario having a configured experiment already setup."""
    name = 'voici_voila'
    config = {'name': name}
    config['algorithms'] = {
        'gradient_descent': {
            'learning_rate': 0.1
            }
        }
    config['pool_size'] = 1
    config['max_trials'] = 100
    config['user_args'] = [
        os.path.abspath(os.path.join(os.path.dirname(__file__), "black_box.py")),
        "-x~uniform(-50, 50, precision=None)"]

    with OrionState():
        experiment = experiment_builder.build_from_args(config)

        workon(experiment, 100, 100, 100, 100, 100)

        storage = get_storage()

        exp = list(storage.fetch_experiments({'name': name}))
        assert len(exp) == 1
        exp = exp[0]
        assert '_id' in exp
        assert exp['name'] == name
        assert exp['pool_size'] == 1
        assert exp['max_trials'] == 100
        assert exp['algorithms'] == {'gradient_descent': {'learning_rate': 0.1,
                                                          'dx_tolerance': 1e-7}}
        assert 'user' in exp['metadata']
        assert 'datetime' in exp['metadata']
        assert 'user_script' in exp['metadata']
        assert exp['metadata']['user_args'] == config['user_args']

        trials = list(storage.fetch_trials(experiment))
        assert len(trials) <= 15
        trials = list(sorted(trials, key=lambda trial: trial.submit_time))
        assert trials[-1].status == 'completed'
        for result in trials[-1].results:
            assert result.type != 'constraint'
            if result.type == 'objective':
                assert abs(result.value - 23.4) < 1e-6
                assert result.name == 'example_objective'
            elif result.type == 'gradient':
                res = numpy.asarray(result.value)
                assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-7
                assert result.name == 'example_gradient'
        params = trials[-1]._params
        assert len(params) == 1
        assert params[0].name == '/x'
        assert params[0].type == 'real'
        assert (params[0].value - 34.56789) < 1e-5


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_stress_unique_folder_creation(database, monkeypatch, tmpdir, capfd):
    """Test integration with a possible framework that needs to create
    unique directories per trial.
    """
    # XXX: return and complete test when there is a way to control random
    # seed of Oríon
    how_many = 2
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--max-trials={}".format(how_many),
                         "--pool-size=1",
                         "--name=lalala",
                         "--config", "./stress_gradient.yaml",
                         "./dir_per_trial.py",
                         "--dir={}".format(str(tmpdir)),
                         "--other-name", "{exp.name}",
                         "--name", "{trial.hash_name}",
                         "-x~gaussian(30, 10)"])

    exp = list(database.experiments.find({'name': 'lalala'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']

    # For contingent broken trials, which in this test means that a existing
    # directory was attempted to be created, it means that it's not md5 or
    # bad hash creation to blame, but the finite precision of the floating
    # point representation. Specifically, it seems that gradient descent
    # is able to reach such levels of precision jumping around the minimum
    # (notice that an appropriate learning rate was selected in this stress
    # test to create underdamped behaviour), that it begins to suggest same
    # things from the past. This is intended to be shown with the assertions
    # in the for-loop below.
    trials_c = list(database.trials.find({'experiment': exp_id, 'status': 'completed'}))
    list_of_cx = [trial['params'][0]['value'] for trial in trials_c]
    trials_b = list(database.trials.find({'experiment': exp_id, 'status': 'broken'}))
    list_of_bx = [trial['params'][0]['value'] for trial in trials_b]
    for bx in list_of_bx:
        assert bx in list_of_cx

    # ``exp.name`` has been delivered correctly (next 2 assertions)
    assert len(os.listdir(str(tmpdir))) == 1
    # Also, because of the way the demo gradient descent works `how_many` trials
    # can be completed
    assert len(os.listdir(str(tmpdir.join('lalala')))) == how_many
    assert len(trials_c) == how_many
    capfd.readouterr()  # Suppress fd level 1 & 2


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_working_dir_argument_cmdline(database, monkeypatch, tmp_path):
    """Check that a permanent directory is used instead of tmpdir"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    path = str(tmp_path) + "/test"
    assert not os.path.exists(path)
    orion.core.cli.main(["hunt", "-n", "allo", "--working-dir", path,
                         "--max-trials", "2", "--config", "./database_config.yaml",
                         "./black_box.py", "-x~uniform(-50,50)"])

    exp = list(database.experiments.find({'name': 'allo'}))[0]
    assert exp['working_dir'] == path
    assert os.path.exists(path)
    assert os.listdir(path)

    shutil.rmtree(path)


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_tmpdir_is_deleted(database, monkeypatch, tmp_path):
    """Check that temporary directory is deletid tmpdir"""
    tmp_path = os.path.join(tempfile.gettempdir(), 'orion')
    if os.path.exists(tmp_path):
        shutil.rmtree(tmp_path)

    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "-n", "allo", "--max-trials", "2", "--config",
                         "./database_config.yaml", "./black_box.py", "-x~uniform(-50,50)"])

    assert not os.listdir(tmp_path)


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_working_dir_argument_config(database, monkeypatch):
    """Check that workning dir argument is handled properly"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    dir_path = os.path.join('orion', 'test')
    if os.path.exists(dir_path):
        shutil.rmtree(dir_path)

    orion.core.cli.main(["hunt", "-n", "allo", "--max-trials", "2",
                         "--config", "./working_dir_config.yaml", "./black_box.py",
                         "-x~uniform(-50,50)"])

    exp = list(database.experiments.find({'name': 'allo'}))[0]
    assert exp['working_dir'] == dir_path
    assert os.path.exists(dir_path)
    assert os.listdir(dir_path)

    shutil.rmtree(dir_path)


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_run_with_name_only(database, monkeypatch):
    """Test hunt can be executed with experiment name only"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--init-only", "--config", "./orion_config_random.yaml",
                         "./black_box.py", "-x~uniform(-50, 50)"])

    orion.core.cli.main(["hunt", "--max-trials", "20", "--config", "./orion_config_random.yaml"])

    exp = list(database.experiments.find({'name': 'demo_random_search'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 20


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_run_with_name_only_with_trailing_whitespace(database, monkeypatch):
    """Test hunt can be executed with experiment name and trailing whitespace"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["hunt", "--init-only", "--config", "./orion_config_random.yaml",
                         "./black_box.py", "-x~uniform(-50, 50)"])

    orion.core.cli.main(["hunt", "--max-trials", "20",
                         "--config", "./orion_config_random.yaml", ""])

    exp = list(database.experiments.find({'name': 'demo_random_search'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 20


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
@pytest.mark.parametrize("strategy", ['MaxParallelStrategy', 'MeanParallelStrategy'])
def test_run_with_parallel_strategy(database, monkeypatch, strategy):
    """Test hunt can be executed with max parallel strategies"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    with open('strategy_config.yaml') as f:
        config = yaml.safe_load(f.read())

    config_file = '{}_strategy_config.yaml'.format(strategy)

    with open(config_file, 'w') as f:
        config['producer']['strategy'] = strategy
        f.write(yaml.dump(config))

    orion.core.cli.main(["hunt", "--max-trials", "20", "--pool-size", "1",
                         "--config", config_file,
                         "./black_box.py", "-x~uniform(-50, 50)"])

    os.remove(config_file)

    exp = list(database.experiments.find({'name': 'strategy_demo'}))
    assert len(exp) == 1
    exp = exp[0]
    assert exp['producer']['strategy'] == strategy
    print(exp['max_trials'])
    assert '_id' in exp
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 20


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_worker_trials(database, monkeypatch):
    """Test number of trials executed is limited based on worker-trials"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    assert len(list(database.experiments.find({'name': 'demo_random_search'}))) == 0

    orion.core.cli.main(["hunt", "--config", "./orion_config_random.yaml", "--pool-size", "1",
                         "--worker-trials", "0",
                         "./black_box.py", "-x~uniform(-50, 50)"])

    exp = list(database.experiments.find({'name': 'demo_random_search'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']

    assert len(list(database.trials.find({'experiment': exp_id}))) == 0

    # Test only executes 2 trials
    orion.core.cli.main(["hunt", "--name", "demo_random_search", "--worker-trials", "2"])

    assert len(list(database.trials.find({'experiment': exp_id}))) == 2

    # Test only executes 3 more trials
    orion.core.cli.main(["hunt", "--name", "demo_random_search", "--worker-trials", "3"])

    assert len(list(database.trials.find({'experiment': exp_id}))) == 5

    # Test that max-trials has precedence over worker-trials
    orion.core.cli.main(["hunt", "--name", "demo_random_search", "--worker-trials", "5",
                         "--max-trials", "6"])

    assert len(list(database.trials.find({'experiment': exp_id}))) == 6


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_resilience(monkeypatch):
    """Test if Oríon stops after enough broken trials."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    MAX_BROKEN = 3
    orion.core.config.worker.max_broken = 3

    orion.core.cli.main(["hunt", "--config", "./orion_config_random.yaml", "./broken_box.py",
                         "-x~uniform(-50, 50)"])

    exp = experiment_builder.build(name='demo_random_search')
    assert len(exp.fetch_trials_by_status('broken')) == MAX_BROKEN


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_with_shutdown_quickly(monkeypatch):
    """Check simple pipeline with random search is reasonably fast."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    monkeypatch.setattr(orion.core.config.worker, 'heartbeat', 120)

    process = subprocess.Popen(
        ["orion", "hunt", "--config", "./orion_config_random.yaml", "--max-trials", "10",
         "./black_box.py", "-x~uniform(-50, 50)"])

    assert process.wait(timeout=40) == 0


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_with_nondefault_config_keyword(database, monkeypatch):
    """Check that the user script configuration file is correctly used with a new keyword."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.config.worker.user_script_config = 'configuration'
    orion.core.cli.main(["hunt", "--config", "./orion_config_other.yaml",
                         "./black_box_w_config_other.py", "--configuration", "script_config.yaml"])

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    assert len(exp) == 1
    exp = exp[0]
    assert '_id' in exp
    exp_id = exp['_id']
    assert exp['name'] == 'voila_voici'
    assert exp['pool_size'] == 1
    assert exp['max_trials'] == 100
    assert exp['algorithms'] == {'gradient_descent': {'learning_rate': 0.1,
                                                      'dx_tolerance': 1e-7}}
    assert 'user' in exp['metadata']
    assert 'datetime' in exp['metadata']
    assert 'orion_version' in exp['metadata']
    assert 'user_script' in exp['metadata']
    assert exp['metadata']['user_args'] == ['./black_box_w_config_other.py', '--configuration',
                                            'script_config.yaml']

    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) <= 15
    assert trials[-1]['status'] == 'completed'
    trials = list(sorted(trials, key=lambda trial: trial['submit_time']))
    for result in trials[-1]['results']:
        assert result['type'] != 'constraint'
        if result['type'] == 'objective':
            assert abs(result['value'] - 23.4) < 1e-6
            assert result['name'] == 'example_objective'
        elif result['type'] == 'gradient':
            res = numpy.asarray(result['value'])
            assert 0.1 * numpy.sqrt(res.dot(res)) < 1e-7
            assert result['name'] == 'example_gradient'
    params = trials[-1]['params']
    assert len(params) == 1
    assert params[0]['name'] == '/x'
    assert params[0]['type'] == 'real'
    assert (params[0]['value'] - 34.56789) < 1e-5

    orion.core.config.worker.user_script_config = 'config'


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_demo_precision(database, monkeypatch):
    """Test a simple usage scenario."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--max-trials", "2",
                            "./black_box.py"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    trials = list(sorted(trials, key=lambda trial: trial['submit_time']))
    params = trials[-1]['params']
    value = params[0]['value']

    assert value == float(numpy.format_float_scientific(value, precision=4))


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_debug_mode(monkeypatch):
    """Test debug mode."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "--debug", "hunt", "--config", "./orion_config.yaml", "--max-trials", "2",
        "./black_box.py"] + user_args)

    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, EphemeralDB)


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main([])

    captured = capsys.readouterr().out

    assert 'usage:' in captured
    assert 'Traceback' not in captured



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the db commands."""
import pytest

import orion.core.cli


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main(['db'])

    captured = capsys.readouterr().out

    assert 'usage:' in captured
    assert 'Traceback' not in captured



# -*- coding: utf-8 -*-
"""Perform functional tests for the REST endpoint `/`"""


def test_runtime_summary(client):
    """Tests if the instance's meta-summary is present"""
    response = client.simulate_get('/')

    assert response.json['orion']
    assert response.json['server'] == 'gunicorn'
    assert response.json['database'] == 'EphemeralDB'
    assert response.status == "200 OK"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.converter`."""
import os

import pytest

from orion.core.io.convert import (GenericConverter, infer_converter_from_file_type,
                                   JSONConverter, YAMLConverter)


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_infer_a_yaml_converter(yaml_sample_path):
    """Infer a YAMLConverter from file path string."""
    yaml = infer_converter_from_file_type(yaml_sample_path)
    assert isinstance(yaml, YAMLConverter)


def test_infer_a_json_converter(json_sample_path):
    """Infer a JSONConverter from file path string."""
    json = infer_converter_from_file_type(json_sample_path)
    assert isinstance(json, JSONConverter)


def test_unknown_file_ext(unknown_type_sample_path):
    """Check what happens if an unknown config type is given."""
    converter = infer_converter_from_file_type(unknown_type_sample_path)
    assert isinstance(converter, GenericConverter)


@pytest.fixture()
def generic_converter():
    """Return a generic converter."""
    return GenericConverter(expression_prefix='o~')


class TestGenericConverter(object):
    """Test functionality of generic converter of unknown configuration file type."""

    def test_parse(self, unknown_type_sample_path, generic_converter):
        """Test parsing method."""
        ret = generic_converter.parse(unknown_type_sample_path)
        assert ret == {
            'lalala': 'o~uniform(1, 3, shape=(100, 3))',
            'lala': {
                'la': 'o~uniform(1, 3, shape=(100, 3))',
                'l2a': 'o~+gaussian(0, 0.1, shape=(100, 3))'
                },
            '': {
                'lala': {
                    'iela': 'o~uniform(1, 3, shape=(100, 3))',
                    '': {
                        'iela': 'o~uniform(1, 3, shape=(100, 3))',
                        },
                    },
                },
            'aaalispera': 'o~normal(3, 1)',
            'a': 'o~normal(5, 3)',
            'b': 'o~>a_serious_name',
            'a_serious_name': 'o~-',
            'another_serious_name': 'o~loguniform(0.001, 0.5)',
            }

        assert generic_converter.template == \
            """{lalala!s}

{/lala/la!s}
{//lala/iela!s}
{//lala//iela!s}

a:
   b:
    [asdfa~/Iamapath/dont_capture_me,
      ~/Iamapath/dont_capture_me]

yo:
           {aaalispera!s}

[naedw]
a_var={a!s}

~

[naekei:naedw]
other_var = ~
# Rename it who_names_their_variables_a_seriously
a_var = {b!s}

{{'oups':
# remove it
'{a_serious_name!s}',
'iela':'{another_serious_name!s}'
}}

{lala/l2a!s}
"""

    def test_bad_parse_1(self, generic_converter):
        """Check that conflict is reported when duplicates are present."""
        BAD_UNKNOWN_SAMPLE_1 = os.path.join(TEST_DIR, '..', 'bad_config1.txt')
        with pytest.raises(ValueError) as exc:
            generic_converter.parse(BAD_UNKNOWN_SAMPLE_1)
        assert '/lala/la' in str(exc.value)

    def test_bad_parse_2(self, generic_converter):
        """Check that conflict is reported even when more sneaky duplicate happens."""
        BAD_UNKNOWN_SAMPLE_2 = os.path.join(TEST_DIR, '..', 'bad_config2.txt')
        with pytest.raises(ValueError) as exc:
            generic_converter.parse(BAD_UNKNOWN_SAMPLE_2)
        assert 'lala/la' in str(exc.value)

    def test_bad_parse_3(self, generic_converter):
        """Check that conflict is reported if a namespace points is both and not final."""
        BAD_UNKNOWN_SAMPLE_3 = os.path.join(TEST_DIR, '..', 'bad_config3.txt')
        with pytest.raises(ValueError) as exc:
            generic_converter.parse(BAD_UNKNOWN_SAMPLE_3)
        assert 'lala' in str(exc.value)

    def test_bad_parse_4(self, generic_converter):
        """Check that conflict is reported if a namespace points is both and not final."""
        BAD_UNKNOWN_SAMPLE_4 = os.path.join(TEST_DIR, '..', 'bad_config4.txt')
        with pytest.raises(ValueError) as exc:
            generic_converter.parse(BAD_UNKNOWN_SAMPLE_4)
        assert 'lala' in str(exc.value)

    def test_generate(self, tmpdir, generic_converter, unknown_type_sample_path):
        """Check generation of a particular instance of configuration."""
        generic_converter.parse(unknown_type_sample_path)

        output_file = str(tmpdir.join("output.lalal"))
        data = {
            'lalala': 'ispi',
            'lala': {
                'la': 5,
                'l2a': '+gaussian(0, 0.1, shape=(100, 3))',
                },
            '': {
                'lala': {
                    'iela': 1,
                    '': {
                        'iela': 2,
                        },
                    },
                },
            'aaalispera': 3,
            'a': 4,
            'b': 'o~>a_serious_name',
            'a_serious_name': 'o~-',
            'another_serious_name': [1, 3],
            }
        generic_converter.generate(output_file, data)

        with open(output_file) as f:
            out = f.read()
        assert out == \
            """ispi

5
1
2

a:
   b:
    [asdfa~/Iamapath/dont_capture_me,
      ~/Iamapath/dont_capture_me]

yo:
           3

[naedw]
a_var=4

~

[naekei:naedw]
other_var = ~
# Rename it who_names_their_variables_a_seriously
a_var = o~>a_serious_name

{'oups':
# remove it
'o~-',
'iela':'[1, 3]'
}

+gaussian(0, 0.1, shape=(100, 3))
"""

    def test_get_state_dict(self, unknown_type_sample_path, unknown_type_template_path,
                            generic_converter):
        """Test getting state dict."""
        generic_converter.parse(unknown_type_sample_path)
        assert generic_converter.get_state_dict() == {
            'expression_prefix': 'o~',
            'has_leading': {
                '/lala//iela': '/',
                '/lala/iela': '/',
                'lala/la': '/'},
            'regex': '([\\/]?[\\w|\\/|-]+)~([\\+]?.*\\)|\\-|\\>[A-Za-z_]\\w*)',
            'template': open(unknown_type_template_path, 'r').read()}

    def test_set_state_dict(self, tmpdir, generic_converter):
        """Test that set_state_dict sets state properly to generate new config."""
        generic_converter.set_state_dict({
            'expression_prefix': '',
            'has_leading': {
                'voici': '/'},
            'regex': generic_converter.regex.pattern,
            'template': """\
voici = {/voici}
voila = voici + {voila}"""})

        output_file = str(tmpdir.join("output.lalal"))

        generic_converter.generate(output_file, {'/voici': 'me voici', 'voila': 'me voila'})

        with open(output_file) as f:
            out = f.read()

        assert out == """\
voici = me voici
voila = voici + me voila"""



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.base` -- What is a search algorithm, optimizer of a process
==============================================================================

.. module:: base
   :platform: Unix
   :synopsis: Formulation of a general search algorithm with respect to some
      objective.

"""
from abc import (ABCMeta, abstractmethod)
import hashlib
import logging

from orion.core.utils import Factory

log = logging.getLogger(__name__)


def infer_trial_id(point):
    """Compute a hashing of a point"""
    return hashlib.md5(str(list(point)).encode('utf-8')).hexdigest()


# pylint: disable=too-many-public-methods
class BaseAlgorithm(object, metaclass=ABCMeta):
    """Base class describing what an algorithm can do.

    Parameters
    ----------
    space : `orion.algo.space.Space`
       Definition of a problem's parameter space.
    kwargs : dict
       Tunable elements of a particular algorithm, a dictionary from
       hyperparameter names to values.

    Notes
    -----
    We are using the No Free Lunch theorem's [1]_[3]_ formulation of an
    `BaseAlgorithm`.
    We treat it as a part of a procedure which in each iteration suggests a
    sample of the parameter space of the problem as a candidate solution and observes
    the results of its evaluation.

    **Developer Note**: Each algorithm's complete specification, i.e.
    implementation of its methods and parameters of its own, lies in a
    separate concrete algorithm class, which must be an **immediate** subclass of
    `BaseAlgorithm`. [The reason for this is current implementation of `Factory`
    metaclass which uses `BaseAlgorithm.__subclasses__()`.] Second, one must
    declare an algorithm's own parameters (tunable elements which could be set
    by configuration). This is done by passing them to `BaseAlgorithm.__init__`
    by calling Python's super with a `Space` object as a positional argument plus
    algorithm's own parameters as keyword arguments. The keys of the keyword
    arguments passed to `BaseAlgorithm.__init__` are interpreted as the algorithm's
    parameter names. So for example, a subclass could be as simple as this
    (regarding the logistics, not an actual algorithm's implementation):

    Examples
    --------
    .. code-block:: python
       :linenos:
       :emphasize-lines: 7

       from orion.algo.base import BaseAlgorithm
       from orion.algo.space import (Integer, Space)

       class MySimpleAlgo(BaseAlgorithm):

           def __init__(self, space, multiplier=1, another_param="a string param"):
               super().__init__(space, multiplier=multiplier, another_param=another_param)

           def suggest(self, num=1):
               print(self.another_param)
               return list(map(lambda x: tuple(map(lambda y: self.multiplier * y, x)),
                               self.space.sample(num)))

           def observe(self, points, results):
               pass

       dim = Integer('named_param', 'norm', 3, 2, shape=(2, 3))
       s = Space()
       s.register(dim)

       algo = MySimpleAlgo(s, 2, "I am just sampling!")
       algo.suggest()

    References
    ----------
    .. [1] D. H. Wolpert and W. G. Macready, “No Free Lunch Theorems for Optimization,”
       IEEE Transactions on Evolutionary Computation, vol. 1, no. 1, pp. 67–82, Apr. 1997.
    .. [2] W. G. Macready and D. H. Wolpert, “What Makes An Optimization Problem Hard?,”
       Complexity, vol. 1, no. 5, pp. 40–46, 1996.
    .. [3] D. H. Wolpert and W. G. Macready, “No Free Lunch Theorems for Search,”
       Technical Report SFI-TR-95-02-010, Santa Fe Institute, 1995.

    """

    requires = []

    def __init__(self, space, **kwargs):
        log.debug("Creating Algorithm object of %s type with parameters:\n%s",
                  type(self).__name__, kwargs)
        self._trials_info = {}  # Stores Unique Trial -> Result
        self._space = space
        self._param_names = list(kwargs.keys())
        # Instantiate tunable parameters of an algorithm
        for varname, param in kwargs.items():
            # Check if tunable element is another algorithm
            if isinstance(param, dict) and len(param) == 1:
                subalgo_type = list(param)[0]
                subalgo_kwargs = param[subalgo_type]
                if isinstance(subalgo_kwargs, dict):
                    param = OptimizationAlgorithm(subalgo_type,
                                                  space, **subalgo_kwargs)
            elif isinstance(param, str) and \
                    param.lower() in OptimizationAlgorithm.typenames:
                # pylint: disable=too-many-function-args
                param = OptimizationAlgorithm(param, space)
            elif varname == 'seed':
                self.seed_rng(param)

            setattr(self, varname, param)

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.

        .. note:: This methods does nothing if the algorithm is deterministic.
        """
        pass

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return {'_trials_info': self._trials_info}

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self._trials_info = state_dict.get('_trials_info')

    @abstractmethod
    def suggest(self, num=1):
        """Suggest a `num` of new sets of parameters.

        Parameters
        ----------
        num: int, optional
            Number of points to suggest. Defaults to 1.

        Returns
        -------
        list of points or None
            A list of lists representing points suggested by the algorithm. The algorithm may opt
            out if it cannot make a good suggestion at the moment (it may be waiting for other
            trials to complete), in which case it will return None.

        Notes
        -----
        New parameters must be compliant with the problem's domain `orion.algo.space.Space`.

        """
        pass

    def observe(self, points, results):
        """Observe the `results` of the evaluation of the `points` in the
        process defined in user's script.

        Parameters
        ----------
        points : list of tuples of array-likes
           Points from a `orion.algo.space.Space`.
           Evaluated problem parameters by a consumer.
        results : list of dicts
           Contains the result of an evaluation; partial information about the
           black-box function at each point in `params`.

        Result
        ------
        objective : numeric
           Evaluation of this problem's objective function.
        gradient : 1D array-like, optional
           Contains values of the derivatives of the `objective` function
           with respect to `params`.
        constraint : list of numeric, optional
           List of constraints expression evaluation which must be greater
           or equal to zero by the problem's definition.

        """
        for point, result in zip(points, results):
            point_id = infer_trial_id(point)

            if point_id not in self._trials_info:
                self._trials_info[point_id] = (point, result)

    @property
    def is_done(self):
        """Return True, if an algorithm holds that there can be no further improvement.
        By default, the cardinality of the specified search space will be used to check
        if all possible sets of parameters has been tried.
        """
        if len(self._trials_info) >= self.space.cardinality:
            return True

        if len(self._trials_info) >= getattr(self, 'max_trials', float('inf')):
            return True

        return False

    def score(self, point):  # pylint:disable=no-self-use,unused-argument
        """Allow algorithm to evaluate `point` based on a prediction about
        this parameter set's performance.

        By default, return the same score any parameter (no preference).

        :returns: A subjective measure of expected perfomance.
        :rtype: float

        """
        return 0

    def judge(self, point, measurements):  # pylint:disable=no-self-use,unused-argument
        """Inform an algorithm about online `measurements` of a running trial.

        :param point: A tuple which specifies the values of the (hyper)parameters
           used to execute user's script with.

        This method is to be used as a callback in a client-server communication
        between user's script and a orion's worker using a `BaseAlgorithm`.
        Data returned from this method must be serializable and will be used as
        a response to the running environment. Default response is None.

        .. note:: Calling algorithm to `judge` a `point` based on its online
           `measurements` will effectively change a state in the algorithm (like
           a reinforcement learning agent's hidden state or an automatic early
           stopping mechanism's regression), which it may change the value of
           the property `should_suspend`.

        :returns: None or a serializable dictionary containing named data

        """
        return None

    @property
    def should_suspend(self):
        """Allow algorithm to decide whether a particular running trial is still
        worth to complete its evaluation, based on information provided by the
        `judge` method.

        """
        return False

    @property
    def configuration(self):
        """Return tunable elements of this algorithm in a dictionary form
        appropriate for saving.

        """
        dict_form = dict()
        for attrname in self._param_names:
            if attrname.startswith('_'):  # Do not log _space or others in conf
                continue
            attr = getattr(self, attrname)
            if isinstance(attr, BaseAlgorithm):
                attr = attr.configuration
            dict_form[attrname] = attr

        return {self.__class__.__name__.lower(): dict_form}

    @property
    def space(self):
        """Domain of problem associated with this algorithm's instance."""
        return self._space

    @space.setter
    def space(self, space_):
        """Propagate changes in defined space to possibly nested algorithms."""
        self._space = space_
        for attr in self.__dict__.values():
            if isinstance(attr, BaseAlgorithm):
                attr.space = space_


# pylint: disable=too-few-public-methods,abstract-method
class OptimizationAlgorithm(BaseAlgorithm, metaclass=Factory):
    """Class used to inject dependency on an algorithm implementation.

    .. seealso:: `orion.core.utils.Factory` metaclass and `BaseAlgorithm` interface.
    """

    pass



"""Example usage and tests for :mod:`orion.core.utils.flatten`."""
from orion.core.utils.flatten import flatten, unflatten


def test_basic():
    """Test basic functionality of flatten"""
    d = {'a': {'b': 2, 'c': 3}, 'c': {'d': 3, 'e': 4}}
    assert flatten(d) == {'a.b': 2, 'a.c': 3, 'c.d': 3, 'c.e': 4}


def test_handle_double_ref():
    """Test proper handling of double references in dicts"""
    a = {'b': 2, 'c': 3}
    d = {'a': a, 'd': {'e': a}}
    assert flatten(d) == {'a.b': 2, 'a.c': 3, 'd.e.b': 2, 'd.e.c': 3}


def test_unflatten():
    """Test than unflatten(flatten(x)) is idempotent"""
    a = {'b': 2, 'c': 3}
    d = {'a': a, 'd': {'e': a}}
    assert unflatten(flatten(d)) == d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.setup` -- Module running the setup command
===============================================================

.. module:: setup
   :platform: Unix
   :synopsis: Creates a configurarion file for the database.
"""

import logging

from orion.core.cli.db.setup import main


log = logging.getLogger(__name__)
DESCRIPTION = '(DEPRECATED) Use command `orion db setup` instead'


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    setup_parser = parser.add_parser('setup', help=DESCRIPTION, description=DESCRIPTION)

    setup_parser.set_defaults(func=main)

    return setup_parser


# pylint: disable = unused-argument
def wrap_main(args):
    """Build a configuration file."""
    log.warning('Command `orion setup` is deprecated and will be removed in v0.2.0. Use '
                '`orion db setup` instead.')

    main(args)



#!/usr/bin/env python
"""Common fixtures and utils for tests."""

import copy
import datetime
import getpass
import os

import pytest

from orion.algo.space import (Categorical, Integer, Real, Space)
from orion.core.evc import conflicts
from orion.core.io.convert import (JSONConverter, YAMLConverter)
import orion.core.io.experiment_builder as experiment_builder
from orion.core.io.space_builder import DimensionBuilder
import orion.core.utils.backward as backward
from orion.testing import default_datetime, MockDatetime

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
YAML_SAMPLE = os.path.join(TEST_DIR, 'sample_config.yml')
YAML_DIFF_SAMPLE = os.path.join(TEST_DIR, 'sample_config_diff.yml')
JSON_SAMPLE = os.path.join(TEST_DIR, 'sample_config.json')
UNKNOWN_SAMPLE = os.path.join(TEST_DIR, 'sample_config.txt')
UNKNOWN_TEMPLATE = os.path.join(TEST_DIR, 'sample_config_template.txt')


@pytest.fixture(scope='session')
def yaml_sample_path():
    """Return path with a yaml sample file."""
    return os.path.abspath(YAML_SAMPLE)


@pytest.fixture(scope='session')
def yaml_diff_sample_path():
    """Return path with a different yaml sample file."""
    return os.path.abspath(YAML_DIFF_SAMPLE)


@pytest.fixture
def yaml_config(yaml_sample_path):
    """Return a list containing the key and the sample path for a yaml config."""
    return ['--config', yaml_sample_path]


@pytest.fixture
def yaml_diff_config(yaml_diff_sample_path):
    """Return a list containing the key and the sample path for a different yaml config."""
    return ['--config', yaml_diff_sample_path]


@pytest.fixture(scope='session')
def json_sample_path():
    """Return path with a json sample file."""
    return JSON_SAMPLE


@pytest.fixture
def json_config(json_sample_path):
    """Return a list containing the key and the sample path for a json config."""
    return ['--config', json_sample_path]


@pytest.fixture(scope='session')
def unknown_type_sample_path():
    """Return path with a sample file of unknown configuration filetype."""
    return UNKNOWN_SAMPLE


@pytest.fixture(scope='session')
def unknown_type_template_path():
    """Return path with a template file of unknown configuration filetype."""
    return UNKNOWN_TEMPLATE


@pytest.fixture(scope='session')
def some_sample_path():
    """Return path with a sample file of unknown configuration filetype."""
    return os.path.join(TEST_DIR, 'some_sample_config.txt')


@pytest.fixture
def some_sample_config(some_sample_path):
    """Return a list containing the key and the sample path for some config."""
    return ['--config', some_sample_path]


@pytest.fixture(scope='session')
def yaml_converter():
    """Return a yaml converter."""
    return YAMLConverter()


@pytest.fixture(scope='session')
def json_converter():
    """Return a json converter."""
    return JSONConverter()


@pytest.fixture(scope='module')
def space():
    """Construct a simple space with every possible kind of Dimension."""
    space = Space()
    categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
    dim = Categorical('yolo', categories, shape=2)
    space.register(dim)
    dim = Integer('yolo2', 'uniform', -3, 6)
    space.register(dim)
    dim = Real('yolo3', 'alpha', 0.9)
    space.register(dim)
    return space


@pytest.fixture(scope='module')
def hierarchical_space():
    """Construct a space with hierarchical Dimensions."""
    space = Space()
    categories = {'asdfa': 0.1, 2: 0.2, 3: 0.3, 4: 0.4}
    dim = Categorical('yolo.first', categories, shape=2)
    space.register(dim)
    dim = Integer('yolo.second', 'uniform', -3, 6)
    space.register(dim)
    dim = Real('yoloflat', 'alpha', 0.9)
    space.register(dim)
    return space


@pytest.fixture(scope='module')
def fixed_suggestion():
    """Return the same tuple/sample from a possible space."""
    return (('asdfa', 2), 0, 3.5)


@pytest.fixture()
def with_user_tsirif(monkeypatch):
    """Make ``getpass.getuser()`` return ``'tsirif'``."""
    monkeypatch.setattr(getpass, 'getuser', lambda: 'tsirif')


@pytest.fixture()
def with_user_bouthilx(monkeypatch):
    """Make ``getpass.getuser()`` return ``'bouthilx'``."""
    monkeypatch.setattr(getpass, 'getuser', lambda: 'bouthilx')


@pytest.fixture()
def with_user_dendi(monkeypatch):
    """Make ``getpass.getuser()`` return ``'dendi'``."""
    monkeypatch.setattr(getpass, 'getuser', lambda: 'dendi')


@pytest.fixture()
def random_dt(monkeypatch):
    """Make ``datetime.datetime.utcnow()`` return an arbitrary date."""
    monkeypatch.setattr(datetime, 'datetime', MockDatetime)
    return default_datetime()


@pytest.fixture()
def hacked_exp(with_user_dendi, random_dt, clean_db, create_db_instance):
    """Return an `Experiment` instance with hacked _id to find trials in
    fake database.
    """
    exp = experiment_builder.build(name='supernaedo2-dendi')
    exp._id = 'supernaedo2-dendi'  # white box hack
    return exp


@pytest.fixture()
def trial_id_substitution(with_user_tsirif, random_dt, clean_db, create_db_instance):
    """Replace trial ids by the actual ids of the experiments."""
    db = create_db_instance
    experiments = db.read('experiments', {})
    experiment_dict = dict((experiment['name'], experiment) for experiment in experiments)
    trials = db.read('trials')

    for trial in trials:
        query = {'experiment': trial['experiment']}
        update = {'experiment': experiment_dict[trial['experiment']]['_id']}
        db.write('trials', update, query)


@pytest.fixture()
def refers_id_substitution(with_user_tsirif, random_dt, clean_db, create_db_instance):
    """Replace trial ids by the actual ids of the experiments."""
    db = create_db_instance
    query = {'metadata.user': 'tsirif'}
    selection = {'name': 1, 'refers': 1}
    experiments = db.read('experiments', query, selection)
    experiment_dict = dict((experiment['name'], experiment) for experiment in experiments)

    for experiment in experiments:
        query = {'_id': experiment['_id']}

        root_id = experiment_dict[experiment['refers']['root_id']]['_id']
        if experiment['refers']['parent_id'] is not None:
            parent_id = experiment_dict[experiment['refers']['parent_id']]['_id']
        else:
            parent_id = None
        update = {'refers.root_id': root_id, 'refers.parent_id': parent_id}
        db.write('experiments', update, query)


###
# Fixtures for EVC tests using conflicts, present in both ./evc and ./io.
# Note: Refactoring the EVC out of orion's core should take care of getting those
#       fixtures out of general conftest.py
###


@pytest.fixture
def new_config():
    """Generate a new experiment configuration"""
    user_script = 'abs_path/black_box.py'
    config = dict(
        name='test',
        algorithms='fancy',
        version=1,
        metadata={'VCS': 'to be changed',
                  'user_script': user_script,
                  'user_args': [
                      user_script, '--new~normal(0,2)', '--changed~normal(0,2)'],
                  'user': 'some_user_name'})

    backward.populate_space(config)

    return config


@pytest.fixture
def old_config(create_db_instance):
    """Generate an old experiment configuration"""
    user_script = 'abs_path/black_box.py'
    config = dict(
        name='test',
        algorithms='random',
        version=1,
        metadata={'VCS': {"type": "git",
                          "is_dirty": False,
                          "HEAD_sha": "test",
                          "active_branch": None,
                          "diff_sha": "diff",
                          },
                  'user_script': user_script,
                  'user_args': [
                      user_script, '--missing~uniform(-10,10)', '--changed~uniform(-10,10)'],
                  'user': 'some_user_name'})

    backward.populate_space(config)

    create_db_instance.write('experiments', config)
    return config


@pytest.fixture
def new_dimension_conflict(old_config, new_config):
    """Generate a new dimension conflict for new experiment configuration"""
    name = 'new'
    prior = 'normal(0, 2)'
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def new_dimension_with_default_conflict(old_config, new_config):
    """Generate a new dimension conflict with default value for new experiment configuration"""
    name = 'new'
    prior = 'normal(0, 2, default_value=0.001)'
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def new_dimension_same_prior_conflict(old_config, new_config):
    """Generate a new dimension conflict with different prior for renaming tests"""
    name = 'new'
    prior = 'uniform(-10, 10)'
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def changed_dimension_conflict(old_config, new_config):
    """Generate a changed dimension conflict"""
    name = 'changed'
    old_prior = 'uniform(-10, 10)'
    new_prior = 'normal(0, 2)'
    dimension = DimensionBuilder().build(name, old_prior)
    return conflicts.ChangedDimensionConflict(old_config, new_config,
                                              dimension, old_prior, new_prior)


@pytest.fixture
def missing_dimension_conflict(old_config, new_config):
    """Generate a missing dimension conflict"""
    name = 'missing'
    prior = 'uniform(-10, 10)'
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def missing_dimension_with_default_conflict(old_config, new_config):
    """Generate a missing dimension conflict with a default value"""
    name = 'missing'
    prior = 'uniform(-10, 10, default_value=0.0)'
    dimension = DimensionBuilder().build(name, prior)
    return conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def algorithm_conflict(old_config, new_config):
    """Generate an algorithm configuration conflict"""
    return conflicts.AlgorithmConflict(old_config, new_config)


@pytest.fixture
def code_conflict(old_config, new_config):
    """Generate a code conflict"""
    return conflicts.CodeConflict(old_config, new_config)


@pytest.fixture
def cli_conflict(old_config, new_config):
    """Generate a commandline conflict"""
    new_config = copy.deepcopy(new_config)
    new_config['metadata']['user_args'].append("--some-new=args")
    new_config['metadata']['user_args'].append("--bool-arg")
    backward.populate_space(new_config)
    return conflicts.CommandLineConflict(old_config, new_config)


@pytest.fixture
def config_conflict(old_config, new_config):
    """Generate a script config conflict"""
    return conflicts.ScriptConfigConflict(old_config, new_config)


@pytest.fixture
def experiment_name_conflict(old_config, new_config):
    """Generate an experiment name conflict"""
    return conflicts.ExperimentNameConflict(old_config, new_config)


@pytest.fixture
def bad_exp_parent_config():
    """Generate a new experiment configuration"""
    config = dict(
        _id='test',
        name='test',
        metadata={'user': 'corneauf', 'user_args': ['--x~normal(0,1)']},
        version=1,
        algorithms='random')

    backward.populate_space(config)

    return config


@pytest.fixture
def bad_exp_child_config(bad_exp_parent_config):
    """Generate a new experiment configuration"""
    config = copy.deepcopy(bad_exp_parent_config)
    config['_id'] = "test2"
    config['refers'] = {'parent_id': 'test'}
    config['version'] = 2

    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.evc.experiment`."""

import pytest

from orion.core.evc.adapters import Adapter, CodeChange
from orion.core.evc.experiment import ExperimentNode
from orion.core.worker.experiment import ExperimentView


# To avoid flake8 issues
build_trimmed_tree = "dummy"


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_parent_fetch_trials(create_db_instance):
    """Test that experiment fetch trials from parent properly (adapters are muted)"""
    experiment_name = 'supernaedo2.3'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3']

    experiment = ExperimentView(experiment_name)
    exp_node = ExperimentNode(experiment.name, experiment.version, experiment=experiment)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    assert len(exp_node.children) == 0
    # 2
    # |
    # 2.3
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    for node in exp_node.root:
        node.item._experiment.refers['adapter'] = Adapter.build([])

    query = {'status': 'completed'}
    assert len(experiment.fetch_trials(query)) == 4
    assert len(experiment._experiment._node.parent.item.fetch_trials(query)) == 6
    assert len(experiment.fetch_trials_tree(query)) == 10


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_children_fetch_trials(create_db_instance):
    """Test that experiment fetch trials from children properly (adapters are muted)"""
    experiment_name = 'supernaedo2'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    assert len(exp_node.children) == 1
    # 2
    # |
    # 2.3
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    for node in exp_node.root:
        node.item._experiment.refers['adapter'] = Adapter.build([])

    query = {'status': 'completed'}
    assert len(experiment.fetch_trials(query)) == 6
    assert len(experiment._experiment._node.children[0].item.fetch_trials(query)) == 4
    assert len(experiment.fetch_trials_tree(query)) == 10


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_parent_parent_fetch_trials(create_db_instance):
    """Test that experiment fetch trials from grand parent properly (adapters are muted)"""
    experiment_name = 'supernaedo2.3.1'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.parent.item.name == root_name
    assert len(exp_node.children) == 0
    # 2
    # |
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 3

    experiment.connect_to_version_control_tree(exp_node)

    for node in exp_node.root:
        node.item._experiment.refers['adapter'] = Adapter.build([])

    query = {'status': 'completed'}
    assert len(exp_node.parent.parent.item.fetch_trials(query)) == 6
    assert len(exp_node.parent.item.fetch_trials(query)) == 4
    assert len(exp_node.item.fetch_trials(query)) == 2

    assert len(experiment.fetch_trials_tree(query)) == 6 + 4 + 2


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_children_children_fetch_trials(create_db_instance):
    """Test that experiment fetch trials from grand children properly (adapters are muted)"""
    experiment_name = 'supernaedo2'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].children[0].item.name == leaf_names[0]
    # 2
    # |
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 3

    experiment.connect_to_version_control_tree(exp_node)

    for node in exp_node.root:
        node.item._experiment.refers['adapter'] = Adapter.build([])

    query = {'status': 'completed'}
    assert len(exp_node.item.fetch_trials(query)) == 6
    assert len(exp_node.children[0].item.fetch_trials(query)) == 4
    assert len(exp_node.children[0].children[0].item.fetch_trials(query)) == 2

    assert len(experiment.fetch_trials_tree(query)) == 6 + 4 + 2


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_deletion_adapter_forward(create_db_instance):
    """Test that all decoding_layer=gru pass to children"""
    experiment_name = 'supernaedo2.1'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    assert len(exp_node.children) == 0
    # 2
    # |
    # 2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    assert len(exp_node.item.fetch_trials(query)) == 1
    assert len(exp_node.parent.item.fetch_trials(query)) == 6

    adapter = experiment.refers['adapter']
    assert len(adapter.forward(exp_node.parent.item.fetch_trials(query))) == 1
    assert len(experiment.fetch_trials_tree(query)) == 1 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_deletion_adapter_backward(create_db_instance):
    """Test that all decoding_layer are passed with gru to parent"""
    experiment_name = 'supernaedo2'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2
    # |
    # 2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    assert len(exp_node.item.fetch_trials(query)) == 6
    assert len(exp_node.children[0].item.fetch_trials(query)) == 1

    adapter = exp_node.children[0].item.refers['adapter']
    assert len(adapter.backward(exp_node.children[0].item.fetch_trials(query))) == 1
    assert len(experiment.fetch_trials_tree(query)) == 6 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_renaming_forward(create_db_instance):
    """Test that all encoding_layer are renamed to encoding in children"""
    experiment_name = 'supernaedo2.3'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2
    # |
    # 2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 6
    assert len(exp_node.item.fetch_trials(query)) == 4

    assert all((trial._params[0].name == "/encoding_layer") for trial in parent_trials)

    adapter = experiment.refers['adapter']
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 6
    assert all((trial._params[0].name == "/encoding") for trial in adapted_parent_trials)

    assert len(experiment.fetch_trials_tree(query)) == 6 + 4


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_renaming_backward(create_db_instance):
    """Test that all encoding are renamed to encoding_layer in parent"""
    experiment_name = 'supernaedo2'
    root_name = 'supernaedo2'
    leaf_names = ['supernaedo2.3']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2
    # |
    # 2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 4
    assert len(exp_node.item.fetch_trials(query)) == 6

    assert all((trial._params[0].name == "/encoding") for trial in children_trials)

    adapter = exp_node.children[0].item.refers['adapter']
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 4
    assert all((trial._params[0].name == "/encoding_layer") for trial in adapted_children_trials)

    assert len(experiment.fetch_trials_tree(query)) == 6 + 4


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_prior_change_forward(create_db_instance):
    """Test that trials from parent only pass to children if valid in the new prior"""
    experiment_name = 'supernaedo2.3.1'
    root_name = 'supernaedo2.3'
    leaf_names = ['supernaedo2.3.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 4
    assert len(exp_node.item.fetch_trials(query)) == 2

    adapter = experiment.refers['adapter']
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 1
    assert len(experiment.fetch_trials_tree(query)) == 2 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_prior_change_backward(create_db_instance):
    """Test that all encoding are renamed to encoding_layer in parent"""
    experiment_name = 'supernaedo2.3'
    root_name = 'supernaedo2.3'
    leaf_names = ['supernaedo2.3.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 2
    assert len(exp_node.item.fetch_trials(query)) == 4

    adapter = exp_node.children[0].item.refers['adapter']
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 1

    assert len(experiment.fetch_trials_tree(query)) == 4 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_noeffect_forward(create_db_instance):
    """Test that all trials pass to children when code change type is 'noeffect'"""
    experiment_name = 'supernaedo2.3.1.1'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2.3.1
    # |
    # 2.3.1.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 2
    assert len(exp_node.item.fetch_trials(query)) == 1

    adapter = experiment.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.NOEFFECT
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 2
    assert len(experiment.fetch_trials_tree(query)) == 2 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_noeffect_backward(create_db_instance):
    """Test that all trials pass to parent when code change type is 'noeffect'"""
    experiment_name = 'supernaedo2.3.1'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 1
    assert len(exp_node.item.fetch_trials(query)) == 2

    adapter = exp_node.children[0].item.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.NOEFFECT
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 1

    assert len(experiment.fetch_trials_tree(query)) == 2 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_unsure_forward(create_db_instance):
    """Test that all trials pass to children when code change type is 'unsure'"""
    experiment_name = 'supernaedo2.3.1.2'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.2']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2.3.1
    # |
    # 2.3.1.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 2
    assert len(exp_node.item.fetch_trials(query)) == 1

    adapter = experiment.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.UNSURE
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 2
    assert len(experiment.fetch_trials_tree(query)) == 2 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_unsure_backward(create_db_instance):
    """Test that no trials pass to parent when code change type is 'unsure'"""
    experiment_name = 'supernaedo2.3.1'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.2']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 1
    assert len(exp_node.item.fetch_trials(query)) == 2

    adapter = exp_node.children[0].item.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.UNSURE
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 0

    assert len(experiment.fetch_trials_tree(query)) == 2 + 0


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_break_forward(create_db_instance):
    """Test that no trials pass to children when code change type is 'break'"""
    experiment_name = 'supernaedo2.3.1.3'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.3']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2.3.1
    # |
    # 2.3.1.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 2
    assert len(exp_node.item.fetch_trials(query)) == 1

    adapter = experiment.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.BREAK
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 0
    assert len(experiment.fetch_trials_tree(query)) == 1 + 0


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_code_change_break_backward(create_db_instance):
    """Test that no trials pass to parent when code change type is 'break'"""
    experiment_name = 'supernaedo2.3.1'
    root_name = 'supernaedo2.3.1'
    leaf_names = ['supernaedo2.3.1.3']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2.3
    # |
    # 2.3.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 1
    assert len(exp_node.item.fetch_trials(query)) == 2

    adapter = exp_node.children[0].item.refers['adapter']
    assert adapter.adapters[0].change_type == CodeChange.BREAK
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 0

    assert len(experiment.fetch_trials_tree(query)) == 2 + 0


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_algo_change_forward(create_db_instance):
    """Test that all trials pass to children when algorithm is changed"""
    experiment_name = 'supernaedo2.2.1'
    root_name = 'supernaedo2.2'
    leaf_names = ['supernaedo2.2.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == root_name
    # 2.2
    # |
    # 2.2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    parent_trials = exp_node.parent.item.fetch_trials(query)
    assert len(parent_trials) == 1
    assert len(exp_node.item.fetch_trials(query)) == 1

    adapter = experiment.refers['adapter']
    adapted_parent_trials = adapter.forward(parent_trials)
    assert len(adapted_parent_trials) == 1
    assert len(experiment.fetch_trials_tree(query)) == 1 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_algo_change_backward(create_db_instance):
    """Test that all trials pass to parent when algorithm is changed"""
    experiment_name = 'supernaedo2.2'
    root_name = 'supernaedo2.2'
    leaf_names = ['supernaedo2.2.1']

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == leaf_names[0]
    # 2.2
    # |
    # 2.2.1
    assert len(list(exp_node.root)) == 2

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    children_trials = exp_node.children[0].item.fetch_trials(query)
    assert len(children_trials) == 1
    assert len(exp_node.item.fetch_trials(query)) == 1

    adapter = exp_node.children[0].item.refers['adapter']
    adapted_children_trials = adapter.backward(children_trials)
    assert len(adapted_children_trials) == 1

    assert len(experiment.fetch_trials_tree(query)) == 1 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_full_backward(create_db_instance):
    """Test that trials are adapted properly up to root"""
    experiment_name = 'supernaedo2'
    root_name = None
    leaf_names = []

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.children[0].item.name == 'supernaedo2.1'
    assert exp_node.children[1].item.name == 'supernaedo2.2'
    assert exp_node.children[2].item.name == 'supernaedo2.3'
    assert exp_node.children[1].children[0].item.name == 'supernaedo2.2.1'
    assert exp_node.children[2].children[0].item.name == 'supernaedo2.3.1'
    assert exp_node.children[2].children[0].children[0].item.name == 'supernaedo2.3.1.1'
    assert exp_node.children[2].children[0].children[1].item.name == 'supernaedo2.3.1.2'
    assert exp_node.children[2].children[0].children[2].item.name == 'supernaedo2.3.1.3'
    # 2
    # |    \      \
    # 2.1  2.2    2.3
    #      |      |
    #      2.2.1  2.3.1
    #             |        \         \
    #             2.3.1.1  2.3.1.2  2.3.1.3
    assert len(list(exp_node)) == 9

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    assert len(exp_node.item.fetch_trials(query)) == 6
    assert len(exp_node.children[0].item.fetch_trials(query)) == 1
    assert len(exp_node.children[1].item.fetch_trials(query)) == 1
    assert len(exp_node.children[2].item.fetch_trials(query)) == 4
    assert len(exp_node.children[1].children[0].item.fetch_trials(query)) == 1
    assert len(exp_node.children[2].children[0].item.fetch_trials(query)) == 2
    assert len(exp_node.children[2].children[0].children[0].item.fetch_trials(query)) == 1
    assert len(exp_node.children[2].children[0].children[1].item.fetch_trials(query)) == 1
    assert len(exp_node.children[2].children[0].children[2].item.fetch_trials(query)) == 1

    assert len(experiment.fetch_trials_tree(query)) == 15


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_full_forward(create_db_instance):
    """Test that trials are adapted properly down to leaf"""
    experiment_name = 'supernaedo2.3.1.1'
    root_name = None
    leaf_names = []

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.item.name == 'supernaedo2.3.1'
    assert exp_node.parent.parent.item.name == 'supernaedo2.3'
    assert exp_node.parent.parent.parent.item.name == 'supernaedo2'
    # 2
    # |
    # 2.3
    # |
    # 2.3.1
    # |
    # 2.3.1.1
    assert len(list(exp_node.root)) == 4

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    assert len(exp_node.item.fetch_trials(query)) == 1
    assert len(exp_node.parent.item.fetch_trials(query)) == 2
    assert len(exp_node.parent.parent.item.fetch_trials(query)) == 4
    assert len(exp_node.parent.parent.parent.item.fetch_trials(query)) == 6

    assert len(experiment.fetch_trials_tree(query)) == 2 + 1 + 2 + 1


@pytest.mark.skip(reason="Support for leafs and parents dropped for now")
@pytest.mark.usefixtures("trial_id_substitution")
@pytest.mark.usefixtures("refers_id_substitution")
def test_full_forward_full_backward(create_db_instance):
    """Test that trials are adapted properly forward from parent and backward from leafs"""
    experiment_name = 'supernaedo2.3'
    root_name = 'supernaedo2'
    leaf_names = []

    experiment = ExperimentView(experiment_name)
    exp_node = build_trimmed_tree(experiment, root_name, leaf_names)

    assert exp_node.item.name == experiment_name
    assert exp_node.parent.name == root_name
    assert exp_node.children[0].item.name == 'supernaedo2.3.1'
    assert exp_node.children[0].children[0].item.name == 'supernaedo2.3.1.1'
    assert exp_node.children[0].children[1].item.name == 'supernaedo2.3.1.2'
    assert exp_node.children[0].children[2].item.name == 'supernaedo2.3.1.3'
    # 2
    # |
    # 2.3
    # |
    # 2.3.1
    # |        \        \
    # 2.3.1.1  2.3.1.2  2.3.1.3
    assert len(list(exp_node.root)) == 6

    experiment.connect_to_version_control_tree(exp_node)

    query = {'status': 'completed'}
    assert len(exp_node.parent.item.fetch_trials(query)) == 6
    assert len(exp_node.item.fetch_trials(query)) == 4
    assert len(exp_node.children[0].item.fetch_trials(query)) == 2
    assert len(exp_node.children[0].children[0].item.fetch_trials(query)) == 1
    assert len(exp_node.children[0].children[1].item.fetch_trials(query)) == 1
    assert len(exp_node.children[0].children[2].item.fetch_trials(query)) == 1

    assert len(experiment.fetch_trials_tree(query)) == 6 + 4 + 1 + 1 + 0 + 0



# -*- coding: utf-8 -*-
# pylint:disable=protected-access,too-many-public-methods,too-many-lines
"""
:mod:`orion.core.worker.experiment` -- Description of an optimization attempt
=============================================================================

.. module:: experiment
   :platform: Unix
   :synopsis: Manage history of trials corresponding to a black box process

"""
import copy
import datetime
import logging

import orion.core
from orion.core.evc.adapters import BaseAdapter
from orion.core.evc.experiment import ExperimentNode
from orion.storage.base import FailedUpdate, get_storage, ReadOnlyStorageProtocol

log = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods
class Experiment:
    """Represents an entry in database/experiments collection.

    Attributes
    ----------
    name : str
       Unique identifier for this experiment per `user`.
    id: object
       id of the experiment in the database if experiment is configured. Value is `None`
       if the experiment is not configured.
    refers : dict or list of `Experiment` objects, after initialization is done.
       A dictionary pointing to a past `Experiment` id, ``refers[parent_id]``, whose
       trials we want to add in the history of completed trials we want to re-use.
       For convenience and database effiency purpose, all experiments of a common tree shares
       `refers[root_id]`, with the root experiment refering to itself.
    version: int
        Current version of this experiment.
    metadata : dict
       Contains managerial information about this `Experiment`.
    pool_size : int
       How many workers can participate asynchronously in this `Experiment`.
    max_trials : int
       How many trials must be evaluated, before considering this `Experiment` done.
       This attribute can be updated if the rest of the experiment configuration
       is the same. In that case, if trying to set to an already set experiment,
       it will overwrite the previous one.
    space: Space
       Object representing the optimization space.
    algorithms : `PrimaryAlgo` object.
       Complete specification of the optimization and dynamical procedures taking
       place in this `Experiment`.

    Metadata
    --------
    user : str
       System user currently owning this running process, the one who invoked **Oríon**.
    datetime : `datetime.datetime`
       When was this particular configuration submitted to the database.
    orion_version : str
       Version of **Oríon** which suggested this experiment. `user`'s current
       **Oríon** version.
    user_script : str
       Full absolute path to `user`'s executable.
    user_args : list of str
       Contains separate arguments to be passed when invoking `user_script`,
       possibly templated for **Oríon**.
    user_vcs : str, optional
       User's version control system for this executable's code repository.
    user_version : str, optional
       Current user's repository version.
    user_commit_hash : str, optional
       Current `Experiment`'s commit hash for **Oríon**'s invocation.

    """

    __slots__ = ('name', 'refers', 'metadata', 'pool_size', 'max_trials', 'version',
                 'space', 'algorithms', 'producer', 'working_dir', '_id',
                 '_node', '_storage')
    non_branching_attrs = ('pool_size', 'max_trials')

    def __init__(self, name, version=None):
        self._id = None
        self.name = name
        self.version = version if version else 1
        self._node = None
        self.refers = {}
        self.metadata = {}
        self.pool_size = None
        self.max_trials = None
        self.space = None
        self.algorithms = None
        self.working_dir = None
        self.producer = {}
        # this needs to be an attribute because we override it in ExperienceView
        self._storage = get_storage()

        self._node = ExperimentNode(self.name, self.version, experiment=self)

    def fetch_trials(self, with_evc_tree=False):
        """Fetch all trials of the experiment"""
        return self._select_evc_call(with_evc_tree, 'fetch_trials')

    def get_trial(self, trial=None, uid=None):
        """Fetch a single Trial, see `orion.storage.base.BaseStorage.get_trial`"""
        return self._storage.get_trial(trial, uid)

    def retrieve_result(self, trial, *args, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.retrieve_result`"""
        return self._storage.retrieve_result(trial, *args, **kwargs)

    def set_trial_status(self, *args, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.set_trial_status`"""
        return self._storage.set_trial_status(*args, **kwargs)

    def reserve_trial(self, score_handle=None):
        """Find *new* trials that exist currently in database and select one of
        them based on the highest score return from `score_handle` callable.

        Parameters
        ----------
        score_handle: callable object, optional
            A way to decide which trial out of the *new* ones to
            to pick as *reserved*, defaults to a random choice.
            Deprecated

        Returns
        -------
        Selected `Trial` object, None if could not find any.

        """
        log.debug('reserving trial with (score: %s)', score_handle)

        self.fix_lost_trials()

        selected_trial = self._storage.reserve_trial(self)
        log.debug('reserved trial (trial: %s)', selected_trial)
        return selected_trial

    def fix_lost_trials(self):
        """Find lost trials and set them to interrupted.

        A lost trial is defined as a trial whose heartbeat as not been updated since two times
        the wait time for monitoring. This usually means that the trial is stalling or has been
        interrupted in some way without its status being changed. This functions finds such
        trials and set them as interrupted so they can be launched again.

        """
        trials = self._storage.fetch_lost_trials(self)

        for trial in trials:
            log.debug('Setting lost trial %s status to interrupted...', trial.id)

            try:
                self._storage.set_trial_status(trial, status='interrupted')
                log.debug('success')
            except FailedUpdate:
                log.debug('failed')

    def update_completed_trial(self, trial, results_file=None):
        """Inform database about an evaluated `trial` with results.

        :param trial: Corresponds to a successful evaluation of a particular run.
        :type trial: `Trial`

        .. note::

            Change status from *reserved* to *completed*.

        """
        trial.status = 'completed'
        trial.end_time = datetime.datetime.utcnow()
        self._storage.retrieve_result(trial, results_file)
        # push trial results updates the entire trial status included
        self._storage.push_trial_results(trial)

    def register_lie(self, lying_trial):
        """Register a *fake* trial created by the strategist.

        The main difference between fake trial and orignal ones is the addition of a fake objective
        result, and status being set to completed. The id of the fake trial is different than the id
        of the original trial, but the original id can be computed using the hashcode on parameters
        of the fake trial. See mod:`orion.core.worker.strategy` for more information and the
        Strategist object and generation of fake trials.

        Parameters
        ----------
        trials: `Trial` object
            Fake trial to register in the database

        Raises
        ------
        orion.core.io.database.DuplicateKeyError
            If a trial with the same id already exist in the database. Since the id is computed
            based on a hashing of the trial, this should mean that an identical trial already exist
            in the database.

        """
        lying_trial.status = 'completed'
        lying_trial.end_time = datetime.datetime.utcnow()
        self._storage.register_lie(lying_trial)

    def register_trial(self, trial, status='new'):
        """Register new trial in the database.

        Inform database about *new* suggested trial with specific parameter values. Trials may only
        be registered one at a time to avoid registration of duplicates.

        Parameters
        ----------
        trials: `Trial` object
            Trial to register in the database

        Raises
        ------
        orion.core.io.database.DuplicateKeyError
            If a trial with the same id already exist in the database. Since the id is computed
            based on a hashing of the trial, this should mean that an identical trial already exist
            in the database.

        """
        stamp = datetime.datetime.utcnow()
        trial.experiment = self._id
        trial.status = status
        trial.submit_time = stamp

        self._storage.register_trial(trial)

    def _select_evc_call(self, with_evc_tree, function, *args, **kwargs):
        if self._node is not None and with_evc_tree:
            return getattr(self._node, function)(*args, **kwargs)

        return getattr(self._storage, function)(self, *args, **kwargs)

    def fetch_trials_by_status(self, status, with_evc_tree=False):
        """Fetch all trials with the given status

        Trials are sorted based on `Trial.submit_time`

        :return: list of `Trial` objects
        """
        return self._select_evc_call(with_evc_tree, 'fetch_trials_by_status', status)

    def fetch_noncompleted_trials(self, with_evc_tree=False):
        """Fetch non-completed trials of this `Experiment` instance.

        Trials are sorted based on `Trial.submit_time`

        .. note::

            It will return all non-completed trials, including new, reserved, suspended,
            interrupted and broken ones.

        :return: list of non-completed `Trial` objects
        """
        return self._select_evc_call(with_evc_tree, 'fetch_noncompleted_trials')

    # pylint: disable=invalid-name
    @property
    def id(self):
        """Id of the experiment in the database if configured.

        Value is `None` if the experiment is not configured.
        """
        return self._id

    @property
    def node(self):
        """Node of the experiment in the version control tree.

        Value is `None` if the experiment is not connected to the version control tree.

        .. seealso::

            :py:meth:`orion.core.worker.experiment.Experiment.connect_to_version_control_tree`

        """
        return self._node

    @property
    def is_done(self):
        """Return True, if this experiment is considered to be finished.

        1. Count how many trials have been completed and compare with `max_trials`.
        2. Ask `algorithms` if they consider there is a chance for further improvement, and
           verify is there is any pending trial.

        .. note::

            To be used as a terminating condition in a ``Worker``.

        """
        trials = self.fetch_trials(with_evc_tree=True)
        num_completed_trials = 0
        num_pending_trials = 0
        for trial in trials:
            if trial.status == 'completed':
                num_completed_trials += 1
            elif trial.status in ['new', 'reserved', 'interrupted']:
                num_pending_trials += 1

        return (
            (num_completed_trials >= self.max_trials) or
            (self.algorithms.is_done and num_pending_trials == 0))

    @property
    def is_broken(self):
        """Return True, if this experiment is considered to be broken.

        Count how many trials are broken and return True if that number has reached
        as given threshold.


        """
        num_broken_trials = self._storage.count_broken_trials(self)
        return num_broken_trials >= orion.core.config.worker.max_broken

    @property
    def configuration(self):
        """Return a copy of an `Experiment` configuration as a dictionary."""
        config = dict()
        for attrname in self.__slots__:
            if attrname.startswith('_'):
                continue
            attribute = copy.deepcopy(getattr(self, attrname))
            config[attrname] = attribute
            if attrname in ['algorithms', 'space']:
                config[attrname] = attribute.configuration
            elif attrname == "refers" and isinstance(attribute.get("adapter"), BaseAdapter):
                config[attrname]['adapter'] = config[attrname]['adapter'].configuration
            elif attrname == "producer" and attribute.get("strategy"):
                config[attrname]['strategy'] = config[attrname]['strategy'].configuration

        if self.id is not None:
            config['_id'] = self.id

        return copy.deepcopy(config)

    @property
    def stats(self):
        """Calculate a stats dictionary for this particular experiment.

        Returns
        -------
        stats : dict

        Stats
        -----
        trials_completed : int
           Number of completed trials
        best_trials_id : int
           Unique identifier of the `Trial` object in the database which achieved
           the best known objective result.
        best_evaluation : float
           Evaluation score of the best trial
        start_time : `datetime.datetime`
           When Experiment was first dispatched and started running.
        finish_time : `datetime.datetime`
           When Experiment reached terminating condition and stopped running.
        duration : `datetime.timedelta`
           Elapsed time.

        """
        completed_trials = self.fetch_trials_by_status('completed')

        if not completed_trials:
            return dict()
        stats = dict()
        stats['trials_completed'] = len(completed_trials)
        stats['best_trials_id'] = None
        trial = completed_trials[0]
        stats['best_evaluation'] = trial.objective.value
        stats['best_trials_id'] = trial.id
        stats['start_time'] = self.metadata['datetime']
        stats['finish_time'] = stats['start_time']
        for trial in completed_trials:
            # All trials are going to finish certainly after the start date
            # of the experiment they belong to
            if trial.end_time > stats['finish_time']:  # pylint:disable=no-member
                stats['finish_time'] = trial.end_time
            objective = trial.objective.value
            if objective < stats['best_evaluation']:
                stats['best_evaluation'] = objective
                stats['best_trials_id'] = trial.id
        stats['duration'] = stats['finish_time'] - stats['start_time']

        return stats

    def __repr__(self):
        """Represent the object as a string."""
        return "Experiment(name=%s, metadata.user=%s, version=%s)" % \
            (self.name, self.metadata['user'], self.version)


# pylint: disable=too-few-public-methods
class ExperimentView(object):
    """Non-writable view of an experiment

    .. seealso::

        :py:class:`orion.core.worker.experiment.Experiment` for writable experiments.

    """

    __slots__ = ('_experiment', )

    #                     Attributes
    valid_attributes = (["_id", "name", "refers", "metadata", "pool_size", "max_trials",
                         "version", "space", "working_dir"] +
                        # Properties
                        ["id", "node", "is_done", "is_broken", "algorithms", "stats",
                         "configuration"] +
                        # Methods
                        ["fetch_trials", "fetch_trials_by_status", "get_trial"])

    def __init__(self, experiment):
        self._experiment = experiment
        self._experiment._storage = ReadOnlyStorageProtocol(experiment._storage)

    def __getattr__(self, name):
        """Get attribute only if valid"""
        if name not in self.valid_attributes:
            raise AttributeError("Cannot access attribute %s on view-only experiments." % name)

        return getattr(self._experiment, name)

    def __repr__(self):
        """Represent the object as a string."""
        return "ExperimentView(name=%s, metadata.user=%s, version=%s)" % \
            (self.name, self.metadata['user'], self.version)



# -*- coding: utf-8 -*-
# pylint: disable=redefined-builtin
"""
:mod:`orion.core.io.config` -- Configuration object
===================================================

.. module:: config
   :platform: Unix
   :synopsis: Configuration object to define package configuration


Highly inspired from https://github.com/mila-iqia/blocks/blob/master/blocks/config.py.

"""
import contextlib
import logging
import os

import yaml


logger = logging.getLogger(__name__)


NOT_SET = object()


@contextlib.contextmanager
def _disable_logger(disable=True):
    if disable:
        logger.disabled = True
    yield

    if disable:
        logger.disabled = False


class ConfigurationError(Exception):
    """Error raised when a configuration value is requested but not set."""


def _curate(key):
    return key.replace('-', '_')


class Configuration:
    """Configuration object

    Provides default values configurable at different levels. The configuration object can have
    global default values, which may be overriden by user with yaml configuration files, environment
    variables or by setting directly the values in the configuration object. In order, direct
    definition overrides, environment variables, which overrides yaml configuration, which overrides
    default values in configuration object definition.

    Examples
    --------
    >>> config = Configuration()
    >>> config.add_option('test', str, 'hello', 'TEST_ENV')
    >>> config.test
    'hello'
    >>> config.load_yaml('some_config.yaml')
    >>> config.test
    'in yaml!'
    >>> os.environ['TEST_ENV'] = 'there!'
    >>> config.test
    'there!'
    >>> config.test = 'here'
    >>> config.test
    'here'

    """

    SPECIAL_KEYS = ['_config', '_subconfigs', '_yaml', '_default', '_env_var', '_help',
                    '_deprecated']

    def __init__(self):
        self._config = {}
        self._subconfigs = {}

    def load_yaml(self, path):
        """Load yaml file and set global default configuration

        Parameters
        ----------
        path: str
            Path to the global configuration file.

        Raises
        -------
        ConfigurationError
            If some option in the yaml file does not exist in the config

        """
        with open(path) as f:
            cfg = yaml.safe_load(f)
            if cfg is None:
                return
            self._load_yaml_dict(cfg)

    def _load_yaml_dict(self, config):
        for key in self._config:
            if key not in config:
                continue
            value = config.pop(key)
            default = self[key + '._default']
            deprecated = self[key + '._deprecated']
            logger.debug('Overwritting "%s" default %s with %s', key, default, value)
            self[key + '._yaml'] = value
            if deprecated and deprecated.get('alternative'):
                logger.debug('Overwritting "%s" default %s with %s', key, default, value)
                self[deprecated.get('alternative') + '._yaml'] = value

        for key in self._subconfigs:
            if key not in config:
                continue

            # pylint: disable=protected-access
            self._subconfigs[key]._load_yaml_dict(config.pop(key))

        if config:
            # Make it fail
            self[next(iter(config.keys()))]

    def __getattr__(self, key):
        """Get the value of the option

        Parameters
        ----------
        key: str
            Name of the option

        Returns
        -------
            Value of the option.

        Raises
        -------
        ConfigurationError
            If the option does not exist

        """
        if key == 'config':
            raise AttributeError

        if key not in self._config and key not in self._subconfigs:
            raise ConfigurationError("Configuration does not have an attribute "
                                     "'{}'.".format(key))
        if key in self._subconfigs:
            return self._subconfigs[key]

        config_setting = self._config[key]
        if 'value' in config_setting:
            value = config_setting['value']
        elif ('env_var' in config_setting and
              config_setting['env_var'] in os.environ):
            value = os.environ[config_setting['env_var']]
        elif 'yaml' in config_setting:
            value = config_setting['yaml']
        elif 'default' in config_setting:
            value = config_setting['default']
        else:
            raise ConfigurationError("Configuration not set and no default "
                                     "provided: {}.".format(key))

        if config_setting.get('deprecated'):
            self._deprecate(key)

        return config_setting['type'](value)

    def __setattr__(self, key, value):
        """Set option value or subconfiguration

        Parameters
        ----------
        key: str
            The key or namespace to set the value of the configuration.
            If the configuration has subconfiguration, the key may be
            hierarchical with each level seperated by dots.
            Ex: 'first.second.third'
        value: object or Configuration
            A general object to set an option or a configuration object to set
            a sub configuration.

        Raises
        ------
        TypeError
            - If value is a configuration and an option is already defined for
            given key, or
            - If the value has an invalid type for the given option, or
            - If no option exists for the given key and the value is not a
            configuration object.

        """
        key = _curate(key)
        if key not in self.SPECIAL_KEYS and key in self._config:
            self._validate(key, value)
            self._config[key]['value'] = value
            if self._config[key].get('deprecated'):
                self._deprecate(key, value)

        elif key in ['_config', '_subconfigs']:
            super(Configuration, self).__setattr__(key, value)

        elif key in self._subconfigs:
            raise ValueError('Configuration already contains subconfiguration {}'.format(key))

        elif isinstance(value, Configuration):
            self._subconfigs[key] = value

        else:
            raise TypeError("Can only set {} as a Configuration, not {}. Use add_option to set a "
                            "new option.".format(key, type(value)))

    # pylint: disable=unused-argument
    def _deprecate(self, key, value=NOT_SET):
        deprecate = self._config[key]['deprecated']
        message = "(DEPRECATED) Option `%s` will be removed in %s."
        args = [deprecate.get('name', key), deprecate['version']]
        if 'alternative' in deprecate:
            message += " Use `%s` instead."
            args.append(deprecate['alternative'])

        logger.warning(message, *args)

    def get(self, key, deprecated='warn'):
        """Access value

        Parameters
        ----------
        key: str
            Key to access in the configuration. Similar to config.key.
        deprecated: str, optional
            If 'warn', the access to deprecated options will log a deprecation warning.
            else if 'ignore', no warning will be logged for access to deprecated options.

        """
        with _disable_logger(disable=(deprecated == 'ignore')):
            value = self[key]

        return value

    def _validate(self, key, value):
        """Validate the (key, value) option

        Raises
        ------
        TypeError
            - If the value is a Configuration object while the key is defined as a normal field
              (str, int, bool, etc). It cannot be overwritten by a subconfiguration.
            - If the type of `value` is not valid for `key`.

        """
        if isinstance(value, Configuration):
            raise TypeError("Cannot overwrite option {} with a configuration".format(key))

        try:
            self._config[key]['type'](value)
        except ValueError as e:
            message = "Option {} of type {} cannot be set to {} with type {}".format(
                key, self._config[key]['type'], value, type(value))
            raise TypeError(message) from e

    def __setitem__(self, key, value):
        """Set option value using dict-like syntax

        Parameters
        ----------
        key: str
            The key or namespace to set the value of the configuration.
            If the configuration has subconfiguration, the key may be
            hierarchical with each levels seperated by dots.
            Ex: 'first.second.third'
        value: object
            A general object to set an option.

        """
        keys = list(map(_curate, key.split(".")))

        # Set in current config for special keys
        if len(keys) == 2 and keys[-1] in self.SPECIAL_KEYS:
            key, field = keys
            self._validate(key, value)
            self._config[key][field.lstrip('_')] = value
            if self._config[key].get('deprecated'):
                self._deprecate(key)

        # Set in current configuration
        elif len(keys) == 1:
            setattr(self, keys[0], value)

        # Recursively in sub configurations
        else:
            subconfig = getattr(self, keys[0])
            if subconfig is None:
                raise KeyError("'{}' is not defined in configuration.".format(keys[0]))
            subconfig[".".join(keys[1:])] = value

    def __getitem__(self, key):
        """Get option value using dict-like syntax

        Parameters
        ----------
        key: str
            The key or namespace to set the value of the configuration.
            If the configuration has subconfiguration, the key may be
            hierarchical with each levels seperated by dots.
            Ex: 'first.second.third'

        """
        keys = list(map(_curate, key.split(".")))

        # Recursively in sub configurations
        if len(keys) == 2 and keys[1] in self.SPECIAL_KEYS:
            key_config = self._config.get(keys[0], None)
            if key_config is None:
                raise ConfigurationError("Configuration does not have an attribute "
                                         "'{}'.".format(keys[0]))
            return key_config.get(keys[1][1:], None)
        elif len(keys) > 1:
            subconfig = getattr(self, keys[0])
            if subconfig is None:
                raise ConfigurationError("Configuration does not have an attribute "
                                         "'{}'.".format(key))
            return subconfig[".".join(keys[1:])]
        # Set in current configuration
        else:
            return getattr(self, keys[0])

    def add_option(self, key, option_type, default=NOT_SET, env_var=None, deprecate=None,
                   help=None):
        """Add a configuration setting.

        Parameters
        ----------
        key : str
            The name of the configuration setting. This must be a valid
            Python attribute name i.e. alphanumeric with underscores.
        option_type : function
            A function such as ``float``, ``int`` or ``str`` which takes
            the configuration value and returns an object of the correct
            type.  Note that the values retrieved from environment
            variables are always strings, while those retrieved from the
            YAML file might already be parsed. Hence, the function provided
            here must accept both types of input.
        default : object, optional
            The default configuration to return if not set. By default none
            is set and an error is raised instead.
        env_var : str, optional
            The environment variable name that holds this configuration
            value. If not given, this configuration can only be set in the
            YAML configuration file.
        deprecate: `dict`, optional
            Should define dict(version, alternative), version at which the deprecated option will be
            removed and alternative to use. A deprecation warning will be logged each time this
            option is set by user. The option `name` can be used in addition to `version` and
            `alternative` to provide a different name then the key. This is useful if the key
            is in a subconfiguration and we want the deprecation error message to include the full
            path. This will add (DEPRECATED) at the beginning of the help message.
        help : str, optionial
            Documentation for the option. Can be reused to build documentation
            or to build parsers with help messages.
            Default help message is 'Undocumented'.

        """
        key = _curate(key)
        if key in self._config or key in self._subconfigs:
            raise ValueError('Configuration already contains {}'.format(key))
        self._config[key] = {'type': option_type}
        if env_var is not None:
            self._config[key]['env_var'] = env_var
        if default is not NOT_SET:
            self._config[key]['default'] = default
        if deprecate is not None:
            if 'version' not in deprecate:
                raise ValueError(f'`version` is missing in deprecate option: {deprecate}')
            self._config[key]['deprecated'] = deprecate

        if help is None:
            help = 'Undocumented'

        if default is not NOT_SET:
            help += ' (default: {})'.format(default)
        if deprecate is not None:
            help = '(DEPRECATED) ' + help
        self._config[key]['help'] = help

    def help(self, key):
        """Return the help message for the given option."""
        return self[key + '._help']

    def add_arguments(self, parser, rename=None):
        """Add arguments to an `argparse` parser based on configuration

        This does not support subconfigurations. They will be ignored.

        Parameters
        ----------
        parser: `argparse.ArgumentParser`
            Parser to which this function will add arguments
        rename: dict, optional
            Mappings to provide different commandline names. Ex `{key: --my-arg}`

        """
        if rename is None:
            rename = dict()

        for key in self._config:
            # TODO: Try with list and nargs='*', but it may case issues with
            # nargs=argparse.REMAINDER.
            if self._config[key]['type'] in (dict, list, tuple):
                continue

            # NOTE: Do not set default, if parser.parse_argv().options[key] is None, then code
            # should look to config[key].
            arg_name = rename.get(key, "--{}".format(key.replace('_', '-')))
            parser.add_argument(
                arg_name, type=self._config[key]['type'],
                help=self._config[key].get('help'))

    def __contains__(self, key):
        """Return True if the option is defined."""
        return key in self._config or key in self._subconfigs

    def to_dict(self):
        """Return a dictionary representation of the configuration"""
        config = dict()

        with _disable_logger():
            for key in self._config:
                config[key] = self[key]

            for key in self._subconfigs:
                config[key] = self[key].to_dict()

        return config
